''' Reports Tachometer stastitics to AWS CloudWatch Metrics. '''

import datetime
import logging
from typing import Optional, Dict

import boto3
import botocore

from .timepiece import Tachometer

class CWTachometer(Tachometer):
    ''' Reports Tachometer stastitics to AWS CloudWatch Metrics.

    The IAM policy associated with the Panorama Appplication Role of this app should grant
    the execution of `cloudwatch:PutMetricData` operation.

    :param namespace: The name of the CloudWatch namespace of this custom metrics.
        It can be for example the name of your project.
    :param metric_name: The name of the CloudWatch metrics. This can be for example
        `frame_processing_time`, if you use CWTachometer to measure frame processing
        time statistics.
    :param dimensions: Additional CloudWatch metrics dimensions of this metric. This
        can be for example the device and application identifier.
    :param stats_interval: Report statistics to CloudWatch with this interval. If using
        standard resolution metrics, this should not be less than 1 minute.
    :param executor: If specified, the metrics will be reported asynchronously, using
        this executor.
    :param region: The AWS region of the CloudWatch metrics.
    :param boto3_session: The boto3 session to be used for sending the CloudWatch metrics.
        If left to None, CWTachometer will use the default session. If the default session
        does not have a default region configured, you might get errors.
    :param parent_logger: If you want to connect the logger of this class to a parent,
        specify it here.
    '''

    # pylint: disable=too-many-arguments,too-few-public-methods
    # We could group the CloudWatch related parameters into a group, but why?
    # Also, public methods are inherited from base class

    def __init__(
        self,
        namespace: str,
        metric_name: str,
        dimensions: Optional[Dict[str, str]] = None,
        stats_interval: datetime.timedelta = datetime.timedelta(seconds=60),
        executor: Optional['concurrent.futures.Executor']=None,
        region: Optional[str] = None,
        boto3_session: Optional[boto3.Session] = None,
        parent_logger: Optional[logging.Logger] = None
    ):
        super().__init__(
            stats_callback=self._stats_calback,
            stats_interval=stats_interval,
            executor=executor
        )
        self.logger = (
            logging.getLogger(self.__class__.__name__) if parent_logger is None else
            parent_logger.getChild(self.__class__.__name__)
        )
        self.dimensions = dimensions or {}
        self.namespace = namespace
        self.metric_name = metric_name
        boto3_session = boto3_session or boto3.Session(region_name=region)
        self.cloudwatch = boto3_session.client('cloudwatch')

    def _cw_dimensions(self):
        return [{ 'Name': name, 'Value': value } for name, value in self.dimensions.items()]

    def _stats_calback(self, timestamp, ticker):
        metric_data = {
            'MetricName': self.metric_name,
            'Dimensions': self._cw_dimensions(),
            'Timestamp': timestamp.astimezone(datetime.timezone.utc),
            'StatisticValues': {
                'SampleCount': ticker.len(),
                'Sum': ticker.sum(),
                'Minimum': ticker.min(),
                'Maximum': ticker.max()
            },
            'Unit': 'Seconds'
        }
        try:
            self.cloudwatch.put_metric_data(
                Namespace=self.namespace,
                MetricData=[metric_data]
            )
        except botocore.exceptions.ClientError as error:
            self.logger.warning('Couldn\'t put data for metric %s.%s',
                                self.namespace, self.metric_name)
            self.logger.warning(str(error))
        except AttributeError:
            self.logger.warning('CloudWatch client is not available.')
        return metric_data



''' This module contains the :class:`AutoIdentity` class that provides information about the
application execution environment. '''

import os
import logging
import datetime
from typing import Dict

import boto3

class AutoIdentity:
    ''' AutoIdentity instance queries metadata of the current application instance.

    The IAM policy associated with the `Panorama Application Role`_
    of this app should grant the execution of
    `panorama:ListApplicationInstances`_ operation.

    Args:
        device_region: The AWS region where this Panorama appliance is registered.
        application_instance_id: The application instance id. If left to `None`,
            :class:`AutoIdentity` will try to find the instance id in the environment variable.
        parent_logger: If you want to connect the logger to a parent, specify it here.

    Upon successfully initialization, :class:`AutoIdentity` will fill out the following properties:

    Attributes:
        application_created_time (datetime.datetime): Application deployment time.
        application_instance_id (str): Application instance id.
        application_name (str): Name of this application.
        application_status (str): Health status of this application.
        application_tags (Dict[str, str]): Tags associated with this application.
        application_description (str): The description of this application.
        device_id (str): Device id of the appliance running this application.
        device_name (str): Name of the appliance running this application.

    .. _`Panorama Application Role`:
        https://docs.aws.amazon.com/panorama/latest/dev/permissions-application.html
    .. _`panorama:ListApplicationInstances`:
        https://docs.aws.amazon.com/service-authorization/latest/reference/list_awspanorama.html#awspanorama-actions-as-permissions
    '''

    # pylint: disable=too-many-instance-attributes,too-few-public-methods
    # This class functions as a data class that reads its values from the environment

    def __init__(
        self,
        device_region: str,
        application_instance_id: str = None,
        parent_logger: logging.Logger = None
    ):
        self._logger = (
            logging.getLogger(self.__class__.__name__) if parent_logger is None else
            parent_logger.getChild(self.__class__.__name__)
        )
        self.application_instance_id: str = (
            application_instance_id or os.environ.get('AppGraph_Uid')
        )
        self.application_name: str = None
        self.device_id: str = None
        self.device_name: str = None
        self.application_created_time: datetime.datetime = None
        self.application_status: str = None
        self.application_tags: Dict[str, str]  = None
        self.application_description: str = None
        if not self.application_instance_id:
            self._logger.warning('Could not find application instance id '
                                 'in environment variable "AppGraph_Uid"')
            return
        self._session = boto3.Session(region_name=device_region)
        self._panorama = self._session.client('panorama')
        app_instance_data = self._app_instance_data(self.application_instance_id)
        if not app_instance_data:
            self._logger.warning('Could not find application instance in service response. '
                                f'Check if application_instance_id={self.application_instance_id} '
                                f'and device_region={device_region} parameters are correct.')
            return
        self._config_from_instance_data(app_instance_data)

    def __repr__(self):
        elements = [f'{a}={getattr(self, a)}' for a in dir(self) if not a.startswith('_')]
        return '<AutoIdentity ' + ' '.join(elements) + '>'

    def _config_from_instance_data(self, instance_data):
        self.application_name = instance_data.get('Name')
        self.device_id = instance_data.get('DefaultRuntimeContextDevice')
        self.device_name = instance_data.get('DefaultRuntimeContextDeviceName')
        self.application_created_time = instance_data.get('CreatedTime')
        self.application_status = instance_data.get('HealthStatus')
        self.application_tags = instance_data.get('Tags')
        self.application_description = instance_data.get('Description')

    def _list_app_instances(self, deployed_only=True):
        next_token = None
        while True:
            kwargs = {'StatusFilter': 'DEPLOYMENT_SUCCEEDED'} if deployed_only else {}
            if next_token:
                kwargs['NextToken'] = next_token
            response = self._panorama.list_application_instances(**kwargs)
            for inst in response['ApplicationInstances']:
                yield inst
            if 'NextToken' in response:
                next_token = response['NextToken']
            else:
                break

    def _app_instance_data(self, application_instance_id):
        matches = [inst for inst in self._list_app_instances()
                   if inst.get('ApplicationInstanceId') == application_instance_id]
        return matches[0] if matches else None



''' Utilities for AWS Panorama application development. '''

import functools

def lazy_property(func):
    ''' Caches the return value of a function, and turns it into a property.

    Intended to be used as a function decorator::

        >>> class Foo:
        >>>     @lazy_property
        >>>     def bar(self):
        >>>         print('expensive calculation')
        >>>         return 'bar'
        >>> foo = Foo()
        >>> foo.bar()
        expensive calculation
        'bar'
        >>> foo.bar()
        'bar'
    '''
    attrib_name = '_' + func.__name__
    @property
    @functools.wraps(func)
    def lazy_wrapper(instance, *args, **kwargs):
        if hasattr(instance, attrib_name):
            return getattr(instance, attrib_name)
        value = func(instance, *args, **kwargs)
        object.__setattr__(instance, attrib_name, value)
        return value
    return lazy_wrapper



''' 2D geometry primitives and implementation of some geometric algorithms. '''

from typing import List, Sequence, Tuple
import collections.abc
import dataclasses
from dataclasses import dataclass
import math
from itertools import islice, cycle, groupby
from numbers import Number
import numpy as np

from . import lazy_property

def _issequence(value):
    ''' Returns True if value is a sequence but not a string. '''
    return isinstance(value, collections.abc.Sequence) and not isinstance(value, str)

def _issequence_or_numpy(value):
    return _issequence(value) or isinstance(value, np.ndarray)

class PointMeta(type):
    @classmethod
    def __instancecheck__(cls, instance):
        ''' Any object that has 'x' and 'y' attributes might be considered a Point '''
        return hasattr(instance, 'x') and hasattr(instance, 'y')

@dataclass(frozen=True)
class Point(metaclass=PointMeta):
    ''' A point on the 2D plane.

    Args:
        x (float): The x coordinate of the point
        y (float): The y coordinate of the point
    '''
    x: float
    ''' The x coordinate of the point '''

    y: float
    ''' The y coordinate of the point '''

    @staticmethod
    def ccw(pt1: 'Point', pt2: 'Point', pt3: 'Point') -> bool:
        ''' Determines if the three points form a counterclockwise angle. If two points are
        equal, or the three points are collinear, this method returns `True`.

        Args:
            pt1: The first point
            pt2: The second point
            pt3: The third point

        Returns:
            `True` if the points form a counterclockwise angle
        '''
        d21, d31 = pt2 - pt1, pt3 - pt1
        return d21.x * d31.y >= d31.x * d21.y

    def distance(self, other: 'Point') -> float:
        ''' Calculates the distance between this and an other point.

        Args:
            other: The other point

        Returns:
            The distance between this and an other point.
        '''
        d = self - other
        return math.sqrt(d.x * d.x + d.y * d.y)

    @classmethod
    def _check_arg(cls, arg, method_name):
        if not isinstance(arg, cls):
            raise TypeError(
                f"unsupported operand type(s) for {method_name}: "
                f"'{cls.__name__}' and '{type(arg).__name__}'"
            )

    def __add__(self, other: 'Point') -> 'Point':
        ''' Adds two points as if they were vectors.

        Arg:
            other: the other point

        Returns:
            A new point, the sum of the two points as if they were vectors.
        '''
        Point._check_arg(other, '+')
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other: 'Point') -> 'Point':
        ''' Subtracts two points as if they were vectors.

        Arg:
            other: the other point

        Returns:
            A new point, the difference of the two points as if they were vectors.
        '''
        Point._check_arg(other, '-')
        return Point(self.x - other.x, self.y - other.y)

    def __truediv__(self, divisor: float) -> 'Point':
        ''' Interpret this point as a vector and divide it by a real number.

        Args:
            divisor: a real number

        Returns:
            The original point divided by the number as if it was a vector.
        '''
        return Point(self.x / divisor, self.y / divisor)


    def __getitem__(self, key: int) -> float:
        ''' Returns the first or the second coordinate of this Point.'''
        if not isinstance(key, int):
            raise TypeError('Point indices must be integers.')
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        else:
            raise IndexError(str(key))

    @classmethod
    def from_value(cls, value):
        ''' Deserializes a Point from different formats.

        Supported formats:

         - sequence containing exactly two numbers
         - dictionary containing numbers under 'x' and 'y' keys
         - Point instance (returns the same instance)

        Args:
            value: the value to be converted

        Return: The new Point instance

        Raises: ValueError if the conversion was not successful.
        '''
        if isinstance(value, Point):
            return value
        elif isinstance(value, collections.abc.Mapping) and 'x' in value and 'y' in value:
            return cls(x=value['x'], y=value['y'])
        elif (
            _issequence(value) and
            len(value) == 2 and
            isinstance(value[0], Number) and
            isinstance(value[1], Number)
        ):
            return cls(x=value[0], y=value[1])
        else:
            raise ValueError(f'Could not convert {value} to Point.')


@dataclass(frozen=True)
class Line:
    ''' A line segment.

    Args:
        pt1 (float): The first point of the line segment
        pt2 (float): The second point of the line segment
    '''
    pt1: Point
    ''' The first point of the line segment '''

    pt2: Point
    ''' The second point of the line segment '''

    def __post_init__(self):
        if not isinstance(self.pt1, Point) or not isinstance(self.pt2, Point):
            raise ValueError('Line arguments "pt1" and "pt2" must be Point objects.')

    def intersects(self, other: 'Line') -> bool:
        ''' Determines if this line segment intersects an other one.

        Args:
            other: The other line segment

        Returns:
            `True` if the two segments intersect each other.
        '''
        return (
            Point.ccw(self.pt1, other.pt1, other.pt2) != Point.ccw(self.pt2, other.pt1, other.pt2)
            and
            Point.ccw(self.pt1, self.pt2, other.pt1) != Point.ccw(self.pt1, self.pt2, other.pt2)
        )

    @classmethod
    def from_value(cls, value):
        ''' Deserializes a Line from different formats.

        Supported formats:

         - sequence containing exactly two values that can be deserialized with Point.from_value
         - dictionary containing such Point values under 'pt1' and 'pt2' keys
         - Line instance (returns the same instance)

        Args:
            value: the value to be converted

        Returns: the Line instance

        Raises: ValueError if the value could not be converted to a Line.
        '''
        if isinstance(value, Line):
            return value
        elif _issequence(value) and len(value) == 2:
            return cls(pt1=Point.from_value(value[0]), pt2=Point.from_value(value[1]))
        elif isinstance(value, collections.abc.Mapping) and 'pt1' in value and 'pt2' in value:
            return cls(pt1=Point.from_value(value['pt1']), pt2=Point.from_value(value['pt2']))
        else:
            raise ValueError(f'Could not convert {value} to Line.')

    def __getitem__(self, key: int) -> Point:
        ''' Returns the first or the second point of this Line.

        Args:
            key: the index passed to the bracket operator, must be 0 or 1.

        Returns:
            The first or the second point of this Line.
        '''
        if not isinstance(key, int):
            raise TypeError('Line indices must be integers.')
        if key == 0:
            return self.pt1
        elif key == 1:
            return self.pt2
        else:
            raise IndexError(str(key))


@dataclass(frozen=True)
class Rectangle:
    ''' An axis aligned rectangle.

    Args:
        pt1 (Point): The first corner of the rectangle
        pt2 (Point): The second corner of the rectangle
    '''

    pt1: Point
    ''' The first corner of the rectangle '''

    pt2: Point
    ''' The second corner of the rectangle '''

    def __post_init__(self):
        if not isinstance(self.pt1, Point) or not isinstance(self.pt2, Point):
            raise ValueError('Rectangle arguments "pt1" and "pt2" must be Point objects.')

    def has_inside(self, pt: Point) -> bool:
        ''' Determines if a point is inside this rectangle.

        Args:
            pt: the point

        Returns:
            `True` if the point lies inside this rectangle.
        '''
        return (
            pt.x >= self.pt_min.x and pt.y >= self.pt_min.y and
            pt.x <= self.pt_max.x and pt.y <= self.pt_max.y
        )

    @lazy_property
    def pt_min(self) -> Point:
        ''' The point with the minimum coordinates of this rectangle. '''
        return Point(min(self.pt1.x, self.pt2.x), min(self.pt1.y, self.pt2.y))

    @lazy_property
    def pt_max(self) -> Point:
        ''' The point with the maximum coordinates of this rectangle. '''
        return Point(max(self.pt1.x, self.pt2.x), max(self.pt1.y, self.pt2.y))

    @lazy_property
    def center(self) -> Point:
        ''' The center of the rectangle. '''
        return Point((self.pt_min.x + self.pt_max.x) / 2, (self.pt_min.y + self.pt_max.y) / 2)

    @lazy_property
    def base(self) -> Point:
        ''' Returns the center of the base of the rectangle. '''
        return Point((self.pt_min.x + self.pt_max.x) / 2, self.pt_max.y)

    @lazy_property
    def size(self) -> Tuple[float, float]:
        ''' The width and height of the rectangle. '''
        return self.width, self.height

    @lazy_property
    def width(self) -> float:
        ''' The width of the rectangle. '''
        return self.pt_max.x - self.pt_min.x

    @lazy_property
    def height(self) -> float:
        ''' The height of the rectangle. '''
        return self.pt_max.y - self.pt_min.y

    @lazy_property
    def area(self) -> float:
        ''' The area of the rectangle. '''
        return self.width * self.height

    @lazy_property
    def aspect_ratio(self) -> float:
        ''' The aspect ratio of the rectangle. '''
        return self.width / self.height

    @property
    def top(self) -> float:
        ''' The top edge of the rectangle. '''
        return self.pt_min.y

    @property
    def left(self) -> float:
        ''' The left edge of the rectangle. '''
        return self.pt_min.x

    @property
    def bottom(self) -> float:
        ''' The bottom edge of the rectangle. '''
        return self.pt_max.y

    @property
    def right(self) -> float:
        ''' The right edge of the rectangle. '''
        return self.pt_max.x

    @property
    def tlbr(self) -> Tuple[float, float, float, float]:
        ''' Returns this rectangles coordinates as a top-left-bottom-right tuple. '''
        return self.top, self.left, self.bottom, self.right

    @property
    def tlhw(self) -> Tuple[float, float, float, float]:
        ''' Returns this rectangles coordinates as a top-left-width-height tuple. '''
        return self.top, self.left, self.height, self.width

    @classmethod
    def from_value(cls, value):
        ''' Converts a tuple in the form of ((0.1, 0.2), (0.3, 0.4)) to a Rectangle.

        Args:
            value: the tuple

        Returns: the Rectangle instance

        Raises: ValueError if the tuple could not be converted to a Rectangle.
        '''
        if isinstance(value, Rectangle):
            return value
        elif _issequence(value) and len(value) == 2:
            return cls(pt1=Point.from_value(value[0]), pt2=Point.from_value(value[1]))
        elif isinstance(value, collections.abc.Mapping) and 'pt1' in value and 'pt2' in value:
            return cls(pt1=Point.from_value(value['pt1']), pt2=Point.from_value(value['pt2']))
        else:
            raise ValueError(f'Could not convert {value} to Rectangle.')


    @classmethod
    def from_tlbr(cls, tlbr: Tuple[float, float, float, float]) -> 'Rectangle':
        ''' Converts a top-left-bottom-right sequence of floats to a Rectangle.

        Args:
            tlbr: A top-left-bottom-right sequence of floats.

        Returns: the Rectangle instance

        Raises: ValueError if the sequence could not be converted to a Rectangle.
        '''
        if _issequence_or_numpy(tlbr) and len(tlbr) == 4:
            return cls(pt1=Point(x=tlbr[1], y=tlbr[0]), pt2=Point(x=tlbr[3], y=tlbr[2]))
        else:
            raise ValueError(f'Could not use {tlbr} as top-left-bottom-right sequence.')

    @classmethod
    def from_tlhw(cls, tlhw: Tuple[float, float, float, float]) -> 'Rectangle':
        ''' Converts a top-left-width-height sequence of floats to a Rectangle.

        Args:
            tlbr: A top-left-width-height sequence of floats.

        Returns: the Rectangle instance

        Raises: ValueError if the sequence could not be converted to a Rectangle.
        '''
        if _issequence_or_numpy(tlhw) and len(tlhw) == 4:
            return cls(
                pt1=Point(x=tlhw[1], y=tlhw[0]),
                pt2=Point(x=tlhw[1]+tlhw[3], y=tlhw[0]+tlhw[2])
            )
        else:
            raise ValueError(f'Could not use {tlhw} as top-left-width-height sequence.')

    def __getitem__(self, key: int) -> float:
        ''' Returns the first or the second point of this Rectangle.

        Args:
            key: the index passed to the bracket operator, must be 0 or 1.

        Returns:
            The first or the second point of this Rectangle.
        '''
        if not isinstance(key, int):
            raise TypeError('Point.__getitem__ accepts only integer keys.')
        if key == 0:
            return self.pt_min
        elif key == 1:
            return self.pt_max
        else:
            raise IndexError(str(key))


@dataclass(frozen=True)
class PolyLine:
    ''' A :class:`PolyLine` is a connected series of line segments.

    Args:
        points: the list of the points of the polyline
        closed: `True` if the :class:`PolyLine` is closed
    '''

    points : Sequence[Point]
    ''' The list of the points of the :class:`PolyLine` '''

    closed : bool = True
    ''' `True` if the :class:`PolyLine` is closed. '''

    def __post_init__(self):
        if not isinstance(self.points, collections.abc.Sequence):
            raise ValueError('PolyLine points argument must be a Sequence.')
        if len(self.points) < 2:
            raise ValueError('PolyLine should contain at least two points.')
        for pt in self.points:
            if not isinstance(pt, Point):
                raise ValueError('The elements of PolyLine points argument must be Point objects.')

    @lazy_property
    def lines(self) -> List[Line]:
        ''' The line segments of this :class:`PolyLine` '''
        lines = [Line(start, end) for start, end in zip(self.points, self.points[1:])]
        if self.closed:
            lines.append(Line(self.points[-1], self.points[0]))
        return lines

    @lazy_property
    def boundingbox(self) -> Rectangle:
        ''' The bounding box of this :class:`PolyLine` '''
        minx = min(pt.x for pt in self.points)
        miny = min(pt.y for pt in self.points)
        maxx = max(pt.x for pt in self.points)
        maxy = max(pt.y for pt in self.points)
        return Rectangle(Point(minx, miny), Point(maxx, maxy))

    @lazy_property
    def self_intersects(self) -> bool:
        ''' Determines if this :class:`PolyLine` self-intersects. '''
        return any(
            l1.intersects(l2)
                for idx, l1 in enumerate(self.lines) for l2 in self.lines[idx + 2:]
        )

    @lazy_property
    def is_convex(self) -> bool:
        ''' Determines if the polygon formed from this :class:`PolyLine` is convex.

        The result of this method is undefined for complex (self-intersecting) polygons.

        Returns:
            `True` if the polygon is convex, False otherwise.
        '''
        if len(self.points) < 4:
            return True

        # Iterate over consequitive point triplets
        it0 = self.points
        it1 = islice(cycle(self.points), 1, None)
        it2 = islice(cycle(self.points), 2, None)

        # Check if all angles are ccw, see
        # https://docs.python.org/3/library/itertools.html#itertools-recipes
        group = groupby(
            Point.ccw(pt0, pt1, pt2) for pt0, pt1, pt2 in zip(it0, it1, it2)
        )
        return next(group, True) and not next(group, False)

    def has_inside(self, point: Point) -> bool:
        ''' Determines if a point is inside this closed :class:`PolyLine`.

        This implementation uses the `ray casting algorithm`_.

        .. _`ray casting algorithm`:
           https://en.wikipedia.org/wiki/Point_in_polygon#Ray_casting_algorithm

        Args:
            point: The point

        Returns:
            `True` if the point is inside this closed :class:`PolyLine`.
        '''
        if not self.closed:
            raise ValueError('PolyLine.has_inside works only for closed polylines.')
        if not self.boundingbox.has_inside(point):
            return False
        ray = Line(Point(self.boundingbox.pt_min.x - 0.01, self.boundingbox.pt_min.y), point)
        n_ints = sum(1 if ray.intersects(line) else 0 for line in self.lines)
        return True if n_ints % 2 == 1 else False

    @classmethod
    def from_value(cls, value, closed=True):
        ''' Converts a tuple in the form of ((0.1, 0.2), (0.3, 0.4), ...) to a PolyLine.

        Args:
            value: the tuple
            closed: flags if the newly created PolyLine should be closed or not.

        Returns: the PolyLine instance

        Raises: ValueError if the tuple could not be converted to a PolyLine.
        '''
        if isinstance(value, PolyLine):
            return value
        elif _issequence(value):
            return cls(points=[Point.from_value(pt) for pt in value], closed=closed)
        elif isinstance(value, collections.abc.Mapping) and 'points' in value:
            closed = bool(value.get('closed', closed))
            return cls(points=[Point.from_value(pt) for pt in value['points']], closed=closed)
        else:
            raise ValueError(f'Could not convert {value} to PolyLine')

    def __getitem__(self, key: int) -> Point:
        ''' Returns a single point of this PolyLine.

        Args:
            key: the index passed to the bracket operator.

        Returns:
            A single point of this PolyLine.
        '''
        if not isinstance(key, int):
            raise TypeError('PolyLine indices must be integers.')
        return self.points[key]

    def __len__(self) -> int:
        ''' Returns the number of line segments in this PolyLine.

        Returns:
            The number of line segments.
        '''
        return len(self.points)


''' This module makes it possible to draw annotations on different
backends with an unified API. Currently, you can draw rectangles and labels with
:mod:`~backpack.annotation` on ``panoramasdk.media`` and OpenCV images
(:class:`numpy arrays <numpy.ndarray>`).'''

from .annotation import (
    Annotation, LabelAnnotation, RectAnnotation, LineAnnotation, MarkerAnnotation,
    PolyLineAnnotation, TimestampAnnotation, BoundingBoxAnnotation
)

from .color import Color



from dataclasses import dataclass
from typing import Union, Optional, Sequence, Mapping, OrderedDict
import collections
import hashlib

@dataclass(frozen=True)
class Color:
    ''' A color in the red, blue, green space.

    The color coordinates are integers in the [0; 255] range.

    Args:
        r (int): The red component of the color
        g (int): The green component of the color
        b (int): The blue component of the color
        alpha (float): The alpha channel of transparency, ranged from `0` to `1`
    '''
    r: int
    ''' The red component of the color. '''
    g: int
    ''' The green component of the color. '''
    b: int
    ''' The blue component of the color. '''
    alpha: float = 1.0
    ''' The alpha component of transparency. '''

    @classmethod
    def from_hex(cls, value: Union[str, int]) -> 'Color':
        ''' Creates a color object from its hexadeciman representation.

        Args:
            value: integer or HTML color string

        Returns:
            a new color object
        '''
        if isinstance(value, str):
            value = value.lstrip('#')
            rgb = tuple(int(value[i:i+2], 16) for i in (0, 2, 4))
            return cls(*rgb)
        elif isinstance(value, int):
            rgb = tuple((value & (0xff << (i * 8))) >> (i * 8) for i in (2, 1, 0))
            return cls(*rgb)
        else:
            raise ValueError('Value argument must be str or int')

    @classmethod
    def from_value(cls, value: Union[str, int, Sequence, Mapping, 'Color']):
        ''' Converts an integer (interpreted as 3 bytes hex value), a HTML color string, a
        sequence of 3 or 4 integers, or a dictionary containing 'r', 'g', 'b' and optionally
        'alpha' keys to a Color object.

        Args:
            value: The value to be converted.

        Returns:
            The new Color object.

        Raises:
            ValueError if the conversion was not successful.
        '''
        if isinstance(value, Color):
            return value
        if isinstance(value, (str, int)):
            return cls.from_hex(value)
        elif (
            isinstance(value, collections.abc.Sequence) and
            (len(value) == 3 or len(value) == 4) and
            all(isinstance(e, int if idx < 3 else float) for idx, e in enumerate(value))
        ):
            return cls(*value)
        elif (
            isinstance(value, collections.abc.Mapping) and
            'r' in value and 'g' in value and 'b' in value
        ):
            params = {k: v for k, v in value.items() if k in ('r', 'g', 'b')}
            alpha = value.get('a', value.get('alpha'))
            if alpha is not None:
                params['alpha'] = alpha
            return cls(**params)
        else:
            raise ValueError(f'Could not convert {value} to a Color')

    @classmethod
    def from_id(cls, identifier: int, salt: str = 'salt') -> 'Color':
        ''' Creates a pseudo-random color from an integer identifier.

        For the same identifier and salt the same color will be generated.

        Args:
            identifier: the identifier
            salt: the salt, change this if you want a different color for the same identifier

        Returns:
            A pseudo-random color based on the identifier and the salt.
        '''
        h = hashlib.md5((salt + str(identifier)).encode('utf-8')).digest()
        return Color(h[0], h[1], h[2])

    def brightness(self, brightness: float) -> 'Color':
        ''' Returns a new Color instance with changed brightness.

        Args:
            brightness: The new brightness, if greater than 1, a brighter color will be returned,
                if smaller than 1, a darker color.

        Returns:
            A new color instance with changed brightness.
        '''
        conv = lambda ch: min(255, int(ch * brightness))
        return Color(r=conv(self.r), g=conv(self.g), b=conv(self.b), alpha=self.alpha)


class ColorMap:
    ''' A simply color map implementation. '''

    colors: Mapping[str, Color] = OrderedDict()
    ''' A map of colors by name. Subclasses are supposed to override this argument. '''

    @classmethod
    def from_name(cls, name: str) -> Optional[Color]:
        ''' Returns a color from its name. '''
        return cls.colors.get(name)

    @classmethod
    def as_list(cls) -> Sequence[Color]:
        ''' Returns the colors of this color map as a list. '''
        return list(cls.colors.values())

    @classmethod
    def color_from_id(cls, identifier: int) -> Color:
        colors = cls.as_list()
        return colors[identifier % len(colors)]


class HTMLColors(ColorMap):
    ''' HTML Basic colors as of https://en.wikipedia.org/wiki/Web_colors#HTML_color_names '''

    WHITE   = Color(255, 255, 255)
    SILVER  = Color(192, 192, 192)
    GRAY    = Color(128, 128, 128)
    BLACK   = Color(  0,   0,   0)
    RED     = Color(255,   0,   0)
    MAROON  = Color(128, 128, 128)
    YELLOW  = Color(255, 255,   0)
    OLIVE   = Color(128, 128,   0)
    LIME    = Color(  0, 255,   0)
    GREEN   = Color(  0, 128,   0)
    AQUA    = Color(  0, 255, 255)
    TEAL    = Color(255, 128, 128)
    BLUE    = Color(  0,   0, 255)
    NAVY    = Color(  0,   0, 128)
    FUCHSIA = Color(255,   0, 255)
    PURPLE  = Color(128,   0, 128)

    colors: Mapping[str, Color] = OrderedDict([
        ('white', WHITE),
        ('silver', SILVER),
        ('gray', GRAY),
        ('black', BLACK),
        ('red', RED),
        ('maroon', MAROON),
        ('yellow', YELLOW),
        ('olive', OLIVE),
        ('lime', LIME),
        ('green', GREEN),
        ('acqua', AQUA),
        ('teal', TEAL),
        ('blue', BLUE),
        ('navy', NAVY),
        ('fuchsia', FUCHSIA),
        ('purple', PURPLE)
    ])



''' This module contains time-related utility methods for measuring code execution time
and scheduling tasks in an external event loop. '''

import time
import datetime
import threading
from collections import deque
from itertools import islice
from typing import List, Deque, Optional, Iterator, Dict, Any, Callable, Tuple
from abc import ABC, abstractmethod

from dateutil.tz import tzlocal

def local_now() -> datetime.datetime:
    ''' Returns the current time in local time zone.

    Returns:
        A timezone aware datettime instance in the local time zone.
    '''
    return datetime.datetime.now(tz=tzlocal())

def local_dt(dt: datetime.datetime) -> datetime.datetime:
    ''' Converts the supplied naive datetime to be time zone aware in the local time zone.

    Args:
        dt: The naive datetime instance.

    Returns:
        A timezone aware datettime instance in the local time zone.
    '''
    return dt.astimezone(tz=tzlocal())

def panorama_timestamp_to_datetime(panorama_ts: Tuple[int, int]) -> datetime.datetime:
    ''' Converts panoramasdk.media.time_stamp (seconds, microsececonds)
    tuple to python datetime.

    Args:
        panorama_ts: The Panorama timestamp

    Returns:
        A python datetime instance.
    '''
    sec, microsec = panorama_ts
    return datetime.datetime.fromtimestamp(sec + microsec / 1000000.0)


class BaseTimer(ABC):
    ''' Base class for code execution time measuring timers.

    Args:
        max_intervals: Maximum number of intervals to remember.
    '''

    # Print at most this many intervals in __repr__
    MAX_REPR_INTERVALS = 5

    def __init__(self, max_intervals:int=10):
        self.intervals: Deque[float] = deque(maxlen=max_intervals)

    def min(self) -> float:
        ''' Returns the shortest time interval between the events (in seconds). '''
        return min(self.intervals) if len(self.intervals) > 0 else 0.0

    def max(self) -> float:
        ''' Returns the longest time interval between the events (in seconds). '''
        return max(self.intervals) if len(self.intervals) > 0 else 0.0

    def mean(self) -> float:
        ''' Returns the mean time interval between the events (in seconds). '''
        return self.sum() / self.len() if self.len() > 0 else 0.0

    def sum(self) -> float:
        ''' Returns the sum of the time interval between recorded events (in seconds). '''
        return sum(self.intervals) if len(self.intervals) > 0 else 0.0

    def len(self) -> int:
        ''' Returns the number of recorded events. '''
        return len(self.intervals)

    def freq(self) -> float:
        ''' Returns the mean frequency of the events (in Hertz). '''
        mean = self.mean()
        return 1 / mean if mean > 0 else 0.0

    def reset(self) -> None:
        ''' Resets the timer. '''
        self.intervals.clear()

    def _repr_props(self) -> Iterator[str]:
        if self.intervals:
            iv_list = [f'{iv:.4f}' for iv in islice(self.intervals, self.MAX_REPR_INTERVALS)]
            if len(self.intervals) > self.MAX_REPR_INTERVALS:
                iv_list.append('...')
            yield f'intervals=[{", ".join(iv_list)}]'
            yield f'min={self.min():.4f}'
            yield f'mean={self.mean():.4f}'
            yield f'max={self.max():.4f}'

    def __repr__(self) -> str:
        elems = [self.__class__.__name__] + list(self._repr_props())
        return '<' + ' '.join(elems) + '>'


class Ticker(BaseTimer):

    ''' A performance profiler that measures the time interval between repeatedly
    occuring events.

    Ticker can also calculate basic statistics of the time intervals.

    Example usage::

        ticker = Ticker(max_intervals=5)
        for i in range(10):
            ticker.tick()
            time.sleep(random.random() / 10)
        print(ticker)

    Results::

        <Ticker intervals=[0.0899, 0.0632, 0.0543, 0.0713, 0.0681] min=0.0543 mean=0.0694 max=0.0899>

    :ivar max_intervals: Maximum number of time intervals to be recorded.
        Only the last max_intervals number of intervals will be kept.
    :ivar intervals: The recorded intervals in seconds between the successive events.
    '''

    def __init__(self, max_intervals:int=10):
        super().__init__(max_intervals=max_intervals)
        self._last_tick: Optional[float] = None

    def tick(self) -> None:
        ''' Registers a tick in this Ticker. '''
        now = time.perf_counter()
        if self._last_tick is not None:
            self.intervals.append(now - self._last_tick)
        self._last_tick = now


class StopWatch(BaseTimer):

    ''' A simple performance profiler with context managers.

    There are two ways to use StopWatch: as a context manager, or with the `tick()`
    method. You can use the same StopWatch object in both ways at the same time.

    When used as a context managaer, StopWatch can be used to measure the
    performance of python code using an elegant API based on context manager.
    You can measure nested and serial execution.

    The second way is to measure the average execution of repeating tasks with
    the `tick()` function call. After enough data samples were collected with
    `tick()`, you can calculate the average execution of the task calling
    `mean_tick()`.

    Example usage::

        import time
        with StopWatch('root') as root:
            with root.child('task1', max_ticks=5) as task:
                time.sleep(0.01)
                with task.child('subtask1.1'):
                    time.sleep(0.03)
                with task.child('subtask1.2'):
                    time.sleep(0.07)
                with task.child('subtask1.3'):
                    time.sleep(0.09)
            with root.child('task2') as task:
                time.sleep(0.17)
        print(root)


    Results::

        <StopWatch name=root total_elapsed=0.9222 children=[
            <StopWatch
                name=task1
                total_elapsed=0.7520
                ticks=[0.0501, 0.0601, 0.0701, 0.0802, 0.0902]
                mean_tick=0.0701
                children=[
                    <StopWatch name=subtask1.1 total_elapsed=0.0301>,
                    <StopWatch name=subtask1.2 total_elapsed=0.0701>,
                    <StopWatch name=subtask1.3 total_elapsed=0.0902>
                ]
            >,
            <StopWatch name=task2 total_elapsed=0.1702>
        ]>

    Args:
        name: The name of this StopWatch
        max_intervals: Maximum number of intervals to be recorded.

    Attributes:
        name (str): The name of this StopWatch
        parent (Optional['StopWatch']): The parent StopWatch
        children (Dict[str, 'StopWatch']): The name-indexed dictionary of the children StopWatches
    '''

    def __init__(self, name:str, max_intervals:int=10):
        super().__init__(max_intervals=max_intervals)
        self.name: str = name
        self.parent: Optional['StopWatch'] = None
        self.children: Dict[str, 'StopWatch'] = {}
        self._start: Optional[float] = None

    def child(self, name:str, max_intervals:Optional[int]=None) -> 'StopWatch':
        ''' Creates a new or returns an existing child of this StopWatch.

        Args:
            name: Name of the child StopWatch.
            max_intervals: Maximum number of intervals to be recorded in the
                child. If None, max_intervals of the parent (this object) will be used.
        '''
        if name in self.children:
            return self.children[name]
        if max_intervals is None:
            max_intervals = self.intervals.maxlen
        child = StopWatch(name, max_intervals=max_intervals)
        child.parent = self
        self.children[name] = child
        return child

    def parents(self) -> Iterator['StopWatch']:
        ''' Returns a generator of all parents of this StopWatch. '''
        current = self
        while True:
            current = current.parent
            if not current:
                break
            yield current

    def full_name(self) -> str:
        ''' Returns the fully qualified name of this StopWatch, including
        all parents' name. '''
        family = [self.name]
        family.extend(p.name for p in self.parents())
        return '.'.join(reversed(family))

    def __enter__(self) -> 'StopWatch':
        ''' Context manager entry point returning self.'''
        self._start = time.perf_counter()
        return self

    def __exit__(self, *_) -> None:
        ''' Context manager exit. '''
        self.intervals.append(time.perf_counter() - self._start)
        self._start = None

    @property
    def level(self) -> int:
        ''' Returns the number of parents. '''
        return len(list(self.parents()))

    def _repr_props(self) -> Iterator[str]:
        yield f'name={self.name}'
        for prop in super()._repr_props():
            yield prop

    def __repr__(self) -> str:
        ''' Indented string representation of this StopWatch.'''
        lvl = self.level
        indent = "    " * lvl
        newline = '\n'
        props = list(self._repr_props())
        if self.children:
            children = (repr(c) for c in self.children.values())
            props.append(f'children=[{", ".join(children)}\n{indent}]')
        return f'{newline if lvl > 0 else ""}{indent}<{self.__class__.__name__} {" ".join(props)}>'


class Callback:
    ''' Encapsulates a callback function and its arguments.

    The callback can be optionally called asynchronously.

    :param cb: The callback function to be called
    :param cbargs: Positional arguments of the callback
    :param cbkwargs: Keyword arguments of the callback
    :param executor: If specified, the callback function will be called asynchronously, using
        this executor.
    '''

    # pylint: disable=invalid-name,too-few-public-methods
    # invalid-name disabled for the `cb` parameter. callable is also taken.
    # too-few-public-methods disabled because this is a wrapper class around a callback function.
    def __init__(
        self,
        cb: Callable,
        cbargs: Optional[List[Any]] = None,
        cbkwargs: Optional[Dict[str, Any]] = None,
        executor: Optional['concurrent.futures.Executor'] = None
    ):
        self.cb = cb
        self.cbargs = cbargs or []
        self.cbkwargs = cbkwargs or {}
        self.executor = executor

    def __call__(self):
        if self.executor:
            self.executor.submit(self.cb, *self.cbargs, **self.cbkwargs)
            return None
        return self.cb(*self.cbargs, **self.cbkwargs)


class Schedule(ABC):
    ''' Schedules a task to be called later with the help of an external
    scheduler.

    The external scheduler is expected to call the `tick()` method periodically,
    most likely from an event-loop.

    :param repeating: If this schedule fires repeatedly
    :param callback: The callback to be called when the scheduler fires
    '''

    def __init__(
        self,
        repeating: bool,
        callback: Callback
    ):
        self.repeating = repeating
        self.callback = callback

    def fire(self) -> None:
        ''' Fires the schedule calling the callback.

        :returns: If not using an executor (the callback is called synchronously), `fire()`
            returns the return value of the callback. Otherwise it returns None.
        '''
        return self.callback()

    @abstractmethod
    def tick(self) -> bool: # pylint: disable=no-self-use
        ''' The heartbeat of the schedule to be called periodically.

        :returns: A tuple of (True, callback_return_value) if the schedule was fired,
            otherwise (False, None)
        '''


class AtSchedule(Schedule):
    ''' Schedules a task to be executed only once at a specific time.

    The task will be executed at the next tick after the specified datetime.

    :param at: When to execute the task
    :param args: Positional arguments to be passed to superclass initializer
    :param kwargs: Keyword arguments to be passed to superclass initializer
    '''

    # pylint: disable=invalid-name
    # Disabled for the `at` parameter

    def __init__(self, at: datetime.datetime, *args: Any, **kwargs: Any):
        super().__init__(False, *args, **kwargs)
        self.at_lock = threading.Lock()
        self.fire_lock = threading.Lock()
        self.at = at

    @property
    def at(self) -> datetime.datetime:
        ''' Property accessor for 'at'. '''
        return self._at

    @at.setter
    def at(self, val: datetime.datetime):
        ''' Property setter for 'at'. '''
        with self.at_lock:
            self._at = val
            self._fired = False

    def _do_tick(self):
        now = local_now()
        if not self._fired and self.at is not None and now > self.at:
            res = self.fire()
            self._fired = True
            return (True, res)
        return (False, None)

    def tick(self) -> bool:
        if self.callback.executor:
            with self.fire_lock:
                return self._do_tick()
        return self._do_tick()


class IntervalSchedule(Schedule):
    ''' Schedules a task to be executed at regular time intervals.

    The task will be executed at the first tick and at each tick after
    the specified time interval has passed.

    :param interval: The time interval of the executions
    :param args: Positional arguments to be passed to superclass initializer
    :param kwargs: Keyword arguments to be passed to superclass initializer
    '''

    def __init__(self, interval: datetime.timedelta, *args: Any, **kwargs: Any):
        super().__init__(True, *args, **kwargs)
        self.interval = interval
        self._next_fire = None

    def _set_next_fire(self, now):
        while self._next_fire <= now:
            self._next_fire += self.interval

    def tick(self) -> bool:
        now = local_now()
        if not self._next_fire:
            self._next_fire = now
        if now >= self._next_fire:
            res = (True, self.fire())
        else:
            res = (False, None)
        self._set_next_fire(now)
        return res


class OrdinalSchedule(Schedule):
    ''' Schedules a task to be executed at each nth tick.

    At the first tick the task will not be executed.

    :param ordinal: Execute the task once in every ordinal number of ticks. An OrdinalSchedule
        with zero ordinal will never fire.
    :param args: Positional arguments to be passed to superclass initializer
    :param kwargs: Keyword arguments to be passed to superclass initializer
    '''

    def __init__(self, ordinal: int, *args: Any, **kwargs: Any):
        super().__init__(True, *args, **kwargs)
        if ordinal < 0:
            raise ValueError('ordinal must be greater or equal than zero')
        self.ordinal = ordinal
        self._counter = 0

    def tick(self):
        if self.ordinal == 0:
            return (False, None)
        self._counter += 1
        if self._counter == self.ordinal:
            res = (True, self.fire())
            self._counter = 0
            return res
        return (False, None)


class AlarmClock:
    ''' An alarm clock can be used to bundle different schedules and send them
    the tick event at once.

    :param schedules: The list of the schedules.
    '''

    # pylint: disable=too-few-public-methods
    # It happens that an alarm clock should only tick:
    # real functionality is implemented by wrapped schedules

    def __init__(self, schedules: List[Schedule]=None):
        self.schedules = schedules or []

    def tick(self) -> None:
        ''' The heartbeat of the alarm clock to be called periodically.

        Will forward the tick to the registered schedules.
        '''
        removables = []
        for schedule in self.schedules:
            fired, _ = schedule.tick()
            if fired and not schedule.repeating:
                removables.append(schedule)
        for schedule in removables:
            self.schedules.remove(schedule)


class Tachometer:
    ''' A Tachometer can be used to measure the frequency of recurring events,
    and periodically report statistics about it.

    Call the `tick()` method of the tachometer each time an atomic event happens.
    For example, if you are interested in the stastics of the frame processing
    time of your application, call `tick` method each time you process a new
    frame.

    Tachometer will periodically call a callback function with the following
    signature::

        def stats_callback(timestamp, ticker):
            pass

    passing the timestamp of the last event, as well as the `Ticker` instance that
    collected the events. You can access the `min()`, `max()`, `sum()` (total processing time)
    and the `len()` (number of measurements) methods of the ticker.

    The `tick()` method will return a tuple with these elements:
     - bool: True, if the internal schedule was fired (you can ignore this value)
     - a tuple with these elements:
       - bool: True, if the stats_callback was called
       - the return value of stats_callback if it was called, else None

    :param stats_callback: A callable with the above signature that will be called
        when new statistics is available.
    :param stat_interval: The interval of the statistics calculation
    :param executor: If specified, callback will be called asynchronously using this executor
    '''

    # pylint: disable=too-few-public-methods
    # It happens that a Tachometer should only tick and sometimes call the callback.

    EXPECTED_MAX_FPS = 100

    def __init__(
        self,
        stats_callback: Callable[[datetime.datetime, 'Ticker'], None],
        stats_interval: datetime.timedelta = datetime.timedelta(seconds=60),
        executor: Optional['concurrent.futures.Executor'] = None
    ):
        self.stats_callback = stats_callback
        self.schedule = IntervalSchedule(
            interval=stats_interval,
            callback=Callback(cb=self._calculate_stats, executor=executor)
        )
        ticker_intervals = int(self.EXPECTED_MAX_FPS * stats_interval.total_seconds())
        self.ticker = Ticker(max_intervals=ticker_intervals)

    def tick(self) -> bool:
        ''' Call this method when a recurring event happens.

        :returns: True if the stat_callback was called in this tick.
        '''
        self.ticker.tick()
        return self.schedule.tick()

    def _calculate_stats(self):
        timestamp = local_now()
        if len(self.ticker.intervals) == 0:
            return (False, None)
        res = self.stats_callback(timestamp, self.ticker)
        self.ticker.intervals.clear()
        return (True, res)



import panoramasdk

from .driver import AnnotationDriverBase

from .annotation import (
    MarkerAnnotation, RectAnnotation, LabelAnnotation, LineAnnotation, PolyLineAnnotation
)

class PanoramaMediaAnnotationDriver(AnnotationDriverBase):
    ''' Annotation driver implementation for Panorama media type images.

    You should pass an ``panoramasdk.media`` instance as the context argument of the
    :meth:`~backpack.annotation.AnnotationDriverBase.render()` method.

    :class:`PanoramaMediaAnnotationDriver` currently does not support colors.
    '''

    MARKER_STYLE_TO_STR = {
        MarkerAnnotation.Style.CROSS: '+',
        MarkerAnnotation.Style.TILTED_CROSS: 'x',
        MarkerAnnotation.Style.STAR: '*',
        MarkerAnnotation.Style.DIAMOND: '<>',
        MarkerAnnotation.Style.SQUARE: '||',
        MarkerAnnotation.Style.TRIANGLE_UP: '/\\',
        MarkerAnnotation.Style.TRIANGLE_DOWN: '\\/',
    }

    def add_rect(self, rect_anno: RectAnnotation, context: 'panoramasdk.media') -> None:
        x1, y1 = rect_anno.rect.pt_min
        x2, y2 = rect_anno.rect.pt_max
        context.add_rect(float(x1), float(y1), float(x2), float(y2))

    def add_label(self, label: LabelAnnotation, context: 'panoramasdk.media') -> None:
        x, y = label.point
        context.add_label(label.text, x, y)

    def add_marker(self, marker: MarkerAnnotation, context: 'panoramasdk.media') -> None:
        marker_str = PanoramaMediaAnnotationDriver.MARKER_STYLE_TO_STR.get(marker.style, '.')
        x, y = marker.point
        context.add_label(marker_str, x, y)

    def add_line(self, label: LineAnnotation, context: 'panoramasdk.media') -> None:
        print( # pragma: no cover
            'WARNING: PanoramaMediaAnnotationDriver.add_line is not implemented'
        )

    def add_polyline(self, polyline: PolyLineAnnotation, context: 'panoramasdk.media') -> None:
        print( # pragma: no cover
            'WARNING: PanoramaMediaAnnotationDriver.add_line is not implemented'
        )



''' :class:`SkyLine` streams OpenCV frames to a `GStreamer`_ pipeline.

:class:`SkyLine` itself is an abstract base class that you can not instantiate directly. Instead,
use one of the subclasses derived from :class:`SkyLine` that provide concrete implementation.
For example, :class:`~backpack.kvs.KVSSkyLine` sends frames to AWS Kinesis Video Streams service,
and :class:`~backpack.rtsp.RTSPSkyLine` streams your frames with a built-in RTSP server.

.. _`GStreamer`: https://gstreamer.freedesktop.org
'''

import os
import subprocess
import logging
from collections import OrderedDict
import datetime
from enum import Enum
from typing import Optional
from abc import ABC, abstractmethod

import cv2
from dotenv import find_dotenv, dotenv_values

from .timepiece import Ticker

USE_LAST_VALUE = -999
''' Using this value for dynamic streaming attributes like fps, width and heigth
will cause to use the values from the last streaming session. '''

class SkyLine(ABC):

    ''' Abstract base class for sending OpenCV frames to a remote service using GStreamer.

    :class:`SkyLine` can be used to create programatically a video stream and send it to
    an external video ingestion service supported by GStreamer. Once the :class:`SkyLine`
    instances is configured and the streaming pipeline was opened by calling
    :meth:`start_streaming`, successive frames can be passed to the :meth:`put` method of the
    instance in OpenCV image format (:class:`numpy.ndarray` with a shape of ``(height, width, 3)``,
    BGR channel order, :class:`numpy.uint8` type).

    The frequency of frames (frames per second) as well as the image
    dimensions (width, heigth) are static during the streaming. You can either
    specify these properties upfront in the constructor, or let SkyLine figure out
    these values. In the later case, up to :attr:`FPS_METER_WARMUP_FRAMES`
    frames (by default 100) will be discarded at the begining of the streaming
    and during this period the value of the stream fps, width and height will
    be determined automatically. In all cases you are expected to call the :meth:`put`
    method with the frequency of the :attr:`video_fps` property, and send images
    of (:attr:`video_width`, :attr:`video_height`) size.

    You should also specify ``GST_PLUGIN_PATH`` variable to the folder
    where the kvssink plugin binaries were built, and add the open-source
    dependency folder of kvssink to the ``LD_LIBRARY_PATH`` variable.
    You can define these variables as environment variables, or in a file
    called ``.env`` in the same folder where this file can be found.

    Args:
        parent_logger: If you want to connect the logger of KVS to a parent,
            specify it here.
        gst_log_file: If you want to redirect GStreamer logs to a file, specify
            the full path of the file in this parameter.
        gst_log_level: If you want to override GStreamer log level configuration,
            specify in this parameter.
        dotenv_path: The path of the .env configuration file. If left to None,
            SkyLine will use the default search mechanism of the python-dotenv library
            to search for the .env file (searching in the current and parent folders).

    Attributes:
        video_width (int): The width of the frames in the video stream.
        video_height (int): The height of the frames in the video stream.
        video_fps (float): The number of frames per second sent to the video stream.
        logger (logging.Logger): The logger instance.
    '''

    # pylint: disable=too-many-instance-attributes
    # We could group the stream parameters like fps, width and height into a structure, but why?

    _FRAME_LOG_FREQUENCY = datetime.timedelta(seconds=60)

    # Wait for so many put() requests before starting streaming.
    # During this period the avgerage FPS will be measured and the
    # stream will be initialized with this FPS
    FPS_METER_WARMUP_FRAMES = 100

    class State(Enum):
        ''' States of the :class:`SkyLine`.

        Attributes:
            STOPPED: The :class:`SkyLine` instance is stopped.
            START_WARMUP: The :class:`SkyLine` instance is about to start the warmup period.
            WARMUP: The :class:`SkyLine` instance is measuring frame rate and frame size during
                the warmup period.
            STREAMING: The :class:`SkyLine` instance is streaming. The
                :meth:`~SkyLine.put` method should be called regularly.
            ERROR: The :class:`SkyLine` instance is encountered an error.
        '''
        STOPPED = 0
        START_WARMUP = 1
        WARMUP = 2
        STREAMING = 3
        ERROR = -1

    def __init__(self,
        parent_logger: Optional[logging.Logger] = None,
        gst_log_file: Optional[str] = None,
        gst_log_level: Optional[str] = None,
        dotenv_path: Optional[str] = None
    ):
        self.video_width = None
        self.video_height = None
        self.video_fps = None
        self.logger = (
            logging.getLogger(self.__class__.__name__) if parent_logger is None else
            parent_logger.getChild(self.__class__.__name__)
        )
        if gst_log_file is not None:
            os.environ['GST_DEBUG_FILE'] = gst_log_file
        if gst_log_level is not None:
            os.environ['GST_DEBUG'] = gst_log_level
        self._config_env(dotenv_path)
        self._check_env()
        self._last_log = datetime.datetime.min
        self._fps_meter_warmup_cnt = 0
        self._fps_meter_ticker = Ticker(self.FPS_METER_WARMUP_FRAMES + 1)
        self._last_fps = None
        self._last_width = None
        self._last_height = None
        self._video_writer = None
        self.state = SkyLine.State.STOPPED

    def _check_gst_plugin(self, plugin_name: str) -> bool:
        ''' Checks if a given GStreamer plugin can be correctly loaded.

        :param plugin_name: The name of the GStreamer plugin
        :return: True if the plugin can be loaded by the GStreamer system.
        '''
        try:
            cmd = f'gst-inspect-1.0 {plugin_name} --plugin'
            env = os.environ.copy()
            env['GST_DEBUG'] = '0'
            subprocess.check_output(cmd.split(' '), stderr=subprocess.STDOUT, env=env)
            self.logger.info('"%s" returned no error', cmd)
            return True
        except subprocess.CalledProcessError as error:
            self.logger.warning(
                '"%s" returned error code=%d, output:\n%s',
                cmd, error.returncode, error.output.decode()
            )
            return False

    @abstractmethod
    def _get_pipeline(self, fps: float, width: int, height: int) -> str:
        ''' Returns to GStreamer pipeline definition.

        Implement this method in subclasses and return the GStreamer pipeline
        definition.'''

    def _put_frame(self, frame):
        size = (self.video_width, self.video_height)
        resized = cv2.resize(frame, size, interpolation=cv2.INTER_LINEAR)
        if not self._video_writer or not self._video_writer.isOpened():
            self._frame_log(lambda: self.logger.warning(
                'Tried to write to cv2.VideoWriter but it is not opened'
            ))
            return False
        self._video_writer.write(resized)
        return True

    def _config_env(self, dotenv_path=None):
        os.environ['GST_DEBUG_NO_COLOR'] = '1'
        dotenv_path = dotenv_path or find_dotenv()
        self.logger.info('Loading config variables from %s', dotenv_path)
        if not dotenv_path or not os.path.isfile(dotenv_path):
            self.logger.warning('dotenv configuration file was not found at path: %s', dotenv_path)
            return
        cfg = dotenv_values(dotenv_path=dotenv_path)
        self.logger.info('Loaded env config structure: %s', cfg)
        if 'GST_PLUGIN_PATH' in cfg:
            os.environ['GST_PLUGIN_PATH'] = cfg['GST_PLUGIN_PATH']
        path_elems = []
        if 'LD_LIBRARY_PATH' in os.environ:
            path_elems.extend(os.environ['LD_LIBRARY_PATH'].split(':'))
        if 'LD_LIBRARY_PATH' in cfg:
            path_elems.append(cfg['LD_LIBRARY_PATH'])
        path_elems = list(OrderedDict.fromkeys(path_elems)) # remove duplicates
        path = ':'.join(path_elems)
        path = path.replace('::', ':')
        os.environ['LD_LIBRARY_PATH'] = path
        if 'LD_PRELOAD' in cfg:
            os.environ['LD_PRELOAD'] = cfg['LD_PRELOAD']

    def _check_env(self):
        def _check_var(var_name, warn=True):
            val = os.environ.get(var_name)
            if val:
                self.logger.info('%s=%s', var_name, val)
            elif warn:
                self.logger.warning('%s environment variable is not defined', var_name)

        _check_var('GST_PLUGIN_PATH')
        _check_var('LD_LIBRARY_PATH')
        _check_var('GST_DEBUG', warn=False)
        _check_var('GST_DEBUG_FILE', warn=False)

        self.logger.info('Local time on host: %s', datetime.datetime.now().isoformat())
        self.logger.info('UTC time on host: %s', datetime.datetime.utcnow().isoformat())

    def _open_stream(self, fps, width, height):
        pipeline = self._get_pipeline(fps, width, height)
        self.logger.info('Opening streaming pipeline')
        self._video_writer = cv2.VideoWriter(
            pipeline, cv2.CAP_GSTREAMER, 0, fps, (width, height)
        )
        self.video_fps = fps
        self.video_width = width
        self.video_height = height
        self._last_fps = fps
        self._last_width = width
        self._last_height = height
        return self._video_writer.isOpened()

    def _close_stream(self):
        if self._video_writer:
            self._video_writer.release()
            self._video_writer = None

    def _frame_log(self, log_fn):
        now = datetime.datetime.now()
        if now - self._last_log > self._FRAME_LOG_FREQUENCY:
            log_fn()
            self._last_log = now

    def _start_warmup(self):
        self._fps_meter_ticker.reset()
        self._fps_meter_warmup_cnt = self.FPS_METER_WARMUP_FRAMES

    def _finish_warmup(self, frame):
        fps = self.video_fps or round(self._fps_meter_ticker.freq())
        width = self.video_width or frame.shape[1]
        height = self.video_height or frame.shape[0]
        return fps, width, height

    def _try_open_stream(self, fps, width, height):
        if not self._open_stream(fps, width, height):
            self.logger.warning('Could not open cv2.VideoWriter')
            self.state = SkyLine.State.ERROR
            return False
        self.state = SkyLine.State.STREAMING
        return True

    @property
    def state(self) -> 'SkyLine.State':
        ''' State of the SkyLine. '''
        return self._state

    @state.setter
    def state(self, state: 'SkyLine.State') -> None:
        ''' Set the state of the SkyLine. '''
        self.logger.info('state = %s', state)
        self._state = state

    # Events

    def start_streaming(
        self,
        fps: float = USE_LAST_VALUE,
        width: int = USE_LAST_VALUE,
        height: int = USE_LAST_VALUE
    ) -> None:
        ''' Start the streaming.

        After calling this method, you are exepcted to call the :meth:`put` method at
        regular intervals. The streaming can be stopped and restarted arbitrary times on
        the same :class:`SkyLine` instance.

        You should specify the desired frame rate and frame dimensions. Using :attr:`USE_LAST_VALUE`
        for any of theses attributes will use the value from the last streaming session. If no
        values are found (or the values are explicitly set to ``None``), a warmup session will be
        started.

        Args:
            fps: The declared frame per seconds of the video. Set this to ``None`` to determine
                this value automatically, or :attr:`USE_LAST_VALUE` to use the value from
                the last streaming session.
            width: The declared width of the video. Set this to ``None`` to determine
                this value automatically, or :attr:`USE_LAST_VALUE` to use the value from
                the last streaming session.
            height: The declared heigth of the video. Set this to ``None`` to determine
                this value automatically, or :attr:`USE_LAST_VALUE` to use the value from
                the last streaming session.
        '''
        self.video_fps = self._last_fps if fps == USE_LAST_VALUE else fps
        self.video_width = self._last_width if width == USE_LAST_VALUE else width
        self.video_height = self._last_height if width == USE_LAST_VALUE else height
        if self.state in [SkyLine.State.WARMUP, SkyLine.State.STREAMING]:
            self._close_stream()
        if any(p is None for p in (self.video_fps, self.video_width, self.video_height)):
            self.logger.info('Starting FPS meter warmup.')
            self.state = SkyLine.State.START_WARMUP
        else:
            self.logger.info('No FPS meter warmup needed.')
            self._try_open_stream(self.video_fps, self.video_width, self.video_height)

    def put(
        self,
        frame: 'numpy.ndarray',
    ) -> bool:
        ''' Put a frame to the video stream.

        Args:
            frame: A numpy array of ``(height, width, 3)`` shape and of :class:`numpy.uint8` type.

        Returns:
            ``True`` if the frame was effectively put on the downstream pipeline.
        '''

        if self.state == SkyLine.State.STOPPED:
            # Streamer paused
            return False

        if self.state == SkyLine.State.ERROR:
            self._frame_log(
                lambda: self.logger.warning(
                    '%s.put() was called in %s state', self.__class__.__name__, self.state
                )
            )
            return False

        if self.state == SkyLine.State.START_WARMUP:
            self._start_warmup()
            self.state = SkyLine.State.WARMUP
            return False

        if self.state == SkyLine.State.WARMUP:
            self._fps_meter_ticker.tick()
            self._fps_meter_warmup_cnt -= 1
            if self._fps_meter_warmup_cnt <= 0:
                fps, width, height = self._finish_warmup(frame)
                self.logger.info(
                    'Finished FPS meter warmup. Determined fps=%f, width=%d, height=%d',
                    fps, width, height
                )
                if self._try_open_stream(fps, width, height):
                    return self._put_frame(frame)

            return False

        if self.state == SkyLine.State.STREAMING:
            return self._put_frame(frame)

        # Should not arrive here
        assert False, f'Unhandled SkyLine state {self.state}' # pragma: no cover

    def stop_streaming(self) -> None:
        ''' Stops the streaming.

        Successive calls to :meth:`put` method will silently discard the frames.
        '''
        self._close_stream()
        self.state = SkyLine.State.STOPPED



import logging
from .annotation import Annotation
from typing import Iterable, Optional, Any
from abc import ABC, abstractmethod

from ..geometry import Point
from .annotation import (
    MarkerAnnotation, RectAnnotation, LabelAnnotation, LineAnnotation, PolyLineAnnotation,
    BoundingBoxAnnotation
)

class AnnotationDriverBase(ABC):
    ''' Base class for annotating drawing drivers.

    Annotation drivers provide an unified API to draw annotations on images of different backends.
    All annotation driver should derive from `AnnotationDriverBase`.

    Args:
        parent_logger: If you want to connect the logger of the annotation driver to a parent,
            specify it here.
    '''
    def __init__(self, parent_logger: Optional[logging.Logger] = None):
        self.logger = (
            logging.getLogger(self.__class__.__name__) if parent_logger is None else
            parent_logger.getChild(self.__class__.__name__)
        )

    def render(self,
        annotations: Iterable[Annotation],
        context: Any
    ) -> Any:
        ''' Renders a collection of annotations on a context.

        Args:
            annotations: An iterable collection of annotation type definied in this module.
            context: The context of the backend. Type is implementation-specific.

        Returns:
            The context.
        '''
        for anno in annotations:
            if isinstance(anno, LabelAnnotation):
                self.add_label(anno, context)
            elif isinstance(anno, RectAnnotation):
                self.add_rect(anno, context)
            elif isinstance(anno, MarkerAnnotation):
                self.add_marker(anno, context)
            elif isinstance(anno, LineAnnotation):
                self.add_line(anno, context)
            elif isinstance(anno, PolyLineAnnotation):
                self.add_polyline(anno, context)
            elif isinstance(anno, BoundingBoxAnnotation):
                self.add_bounding_box(anno, context)
            else:
                raise ValueError(f'Unknown annotation type: {type(anno)}')
        return context

    @abstractmethod
    def add_rect(self, rect: RectAnnotation, context: Any) -> None:
        ''' Renders a rectangle on the frame.

        Args:
            rect: A rectangle annotation.
            context: A backend-specific context object that was passed to the :meth:`render` method.
        '''

    @abstractmethod
    def add_label(self, label: LabelAnnotation, context: Any) -> None:
        ''' Renders a label on the frame.

        Args:
            label: A label annotation.
            context: A backend-specific context object that was passed to the :meth:`render` method.
        '''

    @abstractmethod
    def add_marker(self, marker: MarkerAnnotation, context: Any) -> None:
        ''' Renders a marker on the frame.

        Args:
            marker: A marker annotation.
            context: A backend-specific context object that was passed to the :meth:`render` method.
        '''

    @abstractmethod
    def add_line(self, label: LineAnnotation, context: Any) -> None:
        ''' Renders a line on the frame.

        Args:
            line: A line annotation.
            context: A backend-specific context object that was passed to the :meth:`render` method.
        '''

    @abstractmethod
    def add_polyline(self, polyline: PolyLineAnnotation, context: Any) -> None:
        ''' Renders a polyline.

        Args:
            polyline: A polyline annotation.
            context: A backend-specific context object that was passed to the :meth:`render` method.
        '''

    def add_bounding_box(self, bounding_box: BoundingBoxAnnotation, context: Any) -> None:
        ''' Renders a bounding box.

        Args:
            bounding_box: A bounding box annotation.
            context: A backend-specific context object that was passed to the :meth:`render` method.
        '''
        rect = bounding_box.rectangle
        annos = [RectAnnotation(rect=rect, color=bounding_box.color)]
        if bounding_box.top_label:
            annos.append(
                LabelAnnotation(
                    point=rect.pt_min,
                    text=bounding_box.top_label,
                    color=bounding_box.color,
                    horizontal_anchor=LabelAnnotation.HorizontalAnchor.LEFT,
                    vertical_anchor=LabelAnnotation.VerticalAnchor.BOTTOM
                )
            )
        if bounding_box.bottom_label:
            annos.append(
                LabelAnnotation(
                    point=Point(rect.pt_min.x, rect.pt_max.y),
                    text=bounding_box.bottom_label,
                    color=bounding_box.color,
                    horizontal_anchor=LabelAnnotation.HorizontalAnchor.LEFT,
                    vertical_anchor=LabelAnnotation.VerticalAnchor.TOP
                )
            )
        self.render(annos, context)



''' :class:`~backpack.kvs.KVSSkyLine` is a :class:`~backpack.skyline.SkyLine` implementation
that streams the output of the AWS Panorama application to AWS Kinesis Video Streams service.

To use this class you MUST have the following dependencies correctly configured on your system:

 - `GStreamer 1.0`_ installed with standard plugins pack, libav, tools and development libraries
 - `OpenCV 4.2.0`_, compiled with GStreamer support and Python bindings
 - `Amazon Kinesis Video Streams (KVS) Producer SDK`_ compiled with `kvssink GStreamer plugin
   support`_
 - Environment variable ``GST_PLUGIN_PATH`` configured to point to the directory where the compiled
   binaries of KVS Producer SDK GStreamer plugin is placed
 - Environment vatiable ``LD_LIBRARY_PATH`` including the open source third party dependencies
   compiled by KVS Producer SDK
 - `numpy`_
 - `boto3`_

These dependencies can not be easily specified by a ``requirements.txt`` or a Conda environment.
See the example ``Dockerfile`` on how to install these dependencies on your system.

.. _`GStreamer 1.0`: https://gstreamer.freedesktop.org
.. _`OpenCV 4.2.0`: https://opencv.org/opencv-4-2-0/
.. _`Amazon Kinesis Video Streams (KVS) Producer SDK`:
    https://docs.aws.amazon.com/kinesisvideostreams/latest/dg/producer-sdk-cpp.html
.. _`kvssink GStreamer plugin support`:
    https://docs.aws.amazon.com/kinesisvideostreams/latest/dg/producer-sdk-cpp-gstreamer.html
.. _`numpy`: https://numpy.org
.. _`boto3`: https://boto3.amazonaws.com/v1/documentation/api/latest/index.html
'''

import re
import os
import datetime
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import List, Dict, Any
from abc import ABC, abstractmethod

import boto3
from botocore.credentials import RefreshableCredentials

from .skyline import SkyLine
from .timepiece import AtSchedule, Callback, local_now

def _is_refreshable(credentials):
    return isinstance(credentials, RefreshableCredentials)

def _format_time(datetime_):
    return datetime_.strftime('%Y-%m-%dT%H:%M:%SZ')

class KVSSkyLine(SkyLine):

    ''' Sends OpenCV frames to Kinesis Video Streams.

    :class:`KVSSkyLine` can be used to create programatically a video stream and send
    it to AWS Kinesis Video Streams service.

    When initializing the :class:`KVSSkyLine` instance, you should provide the AWS
    credentials that will be used to stream the video in your AWS account.
    The :class:`KVSCredentialsHandler` subclasses implement different ways of passing the
    credentials to the underlying Kinesis Video Stream Producer. In most of the
    cases, :class:`KVSFileCredentialsHandler` with the default arguments should work well,
    as long as your AWS user or the assumed IAM Role have a policy to put data in KVS.

    You can configure the frame width, height and fps auto-detection as described
    in the :class:`~backpack.skyline.SkyLine` class documentation.

    Args:
        stream_region: The AWS region of the Kinesis Video Stream
        stream_name: The name of the Kinesis Video Stream
        credentials_handler: The credentials handler
        args: Positional arguments to be passed to SkyLine initializer.
        kwargs: Keyword arguments to be passed to SkyLine initializer.
    '''

    def __init__(self,
        stream_region: str,
        stream_name: str,
        credentials_handler: 'KVSCredentialsHandler',
        *args: List[Any],
        **kwargs: Dict[str, Any]
    ):
        super().__init__(*args, **kwargs)
        self.stream_region = stream_region
        self.stream_name = stream_name
        self.credentials_handler = credentials_handler
        if not self._check_gst_plugin('kvssink'):
            raise RuntimeError(
                'kvssink GSTreamer plugin was not found. Please check you have installed it '
                'correctly and set the path to it in the GST_PLUGIN_PATH environment variable '
                '(or in .env file)'
            )

    def _get_pipeline(self, fps, width, height):
        credentials_config_str = self.credentials_handler.plugin_config()
        kvs_config_str = ' '.join([
            'storage-size=512',
            f'stream-name="{self.stream_name}"',
            f'aws-region="{self.stream_region}"',
            f'framerate={fps}',
            credentials_config_str
        ])
        pipeline = ' ! '.join([
            'appsrc',
            'videoconvert',
            f'video/x-raw,format=I420,width={width},height={height},framerate={fps}/1',
            'x264enc bframes=0 key-int-max=45 bitrate=500',
            'video/x-h264,stream-format=avc,alignment=au,profile=baseline',
            f'kvssink {kvs_config_str}'
        ])
        pipeline_safe = self.credentials_handler.plugin_config_mask(pipeline)
        self.logger.info(f'GStreamer pipeline definition:\n{pipeline_safe}')
        return pipeline

    def _put_frame(self, *args, **kwargs):
        result = super()._put_frame(*args, **kwargs)
        self.credentials_handler.check_refresh()
        return result


class KVSCredentialsHandler(ABC):
    ''' Abstract base class for credential providers.

    Credential providers provide AWS credentials to Kinesis Video Stream Producer library.

    If no static credentials are provided in the init method, the Credentials
    Handler will figure out the AWS credentials from the boto3 configuration.

    Use of static credentials are recommended at most for testing purposes. If you
    instantiate this class with custom credentials in the init method arguments, or
    calling the init method from a workstation where you've configured an AWS user
    credentials, you will get static credentials. If the caller code assumed an IAM
    role, you will use dynamic credentials.

    When using dynamic credentials, you are expected to call the :meth:`check_refresh`
    method peridocally to control the expiration of the dynamic credentials. Ideally
    you would call this method each time when you want send a new frame to Kinesis
    Video Streams Producer. If the credentials are not expired, this method should add
    almost no overhead.

    You should not instantiate this class: use one if its subclasses instead.

    Args:
        aws_access_key_id: If you want to use custom static credentials, specify your
            AWS access key ID here. This method is not recommended for production builds.
        aws_secret_access_key: If you want to use custom static credentials, specify your
            AWS secret access key here. This method is not recommended for production builds.
        executor: Credential refresh operator will be dispatched to this executor.
        parent_logger: Connect the logger of this class to a parent logger.
    '''

    REFRESH_BEFORE_EXPIRATION = datetime.timedelta(seconds=2 * 60)
    ''' Refresh credentials so many seconds before expiration. '''

    def __init__(
        self,
        aws_access_key_id: str=None,
        aws_secret_access_key: str=None,
        executor: 'concurrent.futures.Executor'=ThreadPoolExecutor(max_workers=1),
        parent_logger: logging.Logger=None
    ):
        self.logger = (
            logging.getLogger(self.__class__.__name__) if parent_logger is None else
            parent_logger.getChild(self.__class__.__name__)
        )
        if aws_access_key_id or aws_secret_access_key:
            self.session = boto3.Session(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_access_key_id
            )
        else:
            self.session = boto3.Session()

        self.caller_arn = self.session.client('sts').get_caller_identity()['Arn']
        self.credentials = self.session.get_credentials()
        self.schedule = AtSchedule(
            at=None,
            callback=Callback(cb=self._refresh, executor=executor)
        )
        self._refresh()

    def _refresh(self):
        # pylint: disable=protected-access
        # since botocore credentials do not give access to expiry date,
        # we've to use protected members
        if _is_refreshable(self.credentials):
            self.logger.info(f'Refreshing credentials using {self.caller_arn}')
            # This will refresh the credentials if neeeded
            advisory = self.credentials.refresh_needed(self.credentials._advisory_refresh_timeout)
            mandatory = self.credentials.refresh_needed(self.credentials._mandatory_refresh_timeout)
            self.logger.info(f'Refresh needed: advisory={advisory}, mandatory={mandatory}')
            self.logger.info(f'Forcing refresh credentials with method '
                             f'{self.credentials._refresh_using}')
            self.credentials._protected_refresh(is_mandatory=True)
            frozen_credentials = self.credentials.get_frozen_credentials()
            expiry_time = self.credentials._expiry_time
            self.logger.info('Got credentials: '
                             f'access_key={frozen_credentials.access_key[:5]}..., '
                             f'secret_key={frozen_credentials.secret_key[:5]}..., '
                             f'token={frozen_credentials.token[:5]}..., '
                             f'expiry={_format_time(expiry_time)}'
            )

            # Schedule next refresh when we're in the mandatory refresh timeout
            # of the underlying credentials
            #
            #                          token expiry_time+
            #                                           |
            #  advisory_refresh_timeout+                |
            #                          |<---- 15m ----->|
            #                          |                |
            #      mandatory_refresh_timeout+           |
            # now                      |    |<-- 10m -->|
            #  v                       v    v           v
            #  -----------------------------------------|
            #                                 ^
            #                   update_timeout+
            #
            next_update = expiry_time - self.REFRESH_BEFORE_EXPIRATION
            credentials = frozen_credentials
            self.logger.info(f'Next update: {_format_time(next_update)}')
            now = local_now()
            if next_update < now:
                msg = ('Next update time is in the past! '
                    f'current_time={_format_time(now)}, '
                    f'next_update={_format_time(next_update)}, '
                    f'expiry={_format_time(expiry_time)}'
                )
                self.logger.warning(msg)
        else:
            self.logger.info('Credentials are static')
            next_update = None
            credentials = self.credentials
        self._save_credentials(credentials, next_update)
        self.schedule.at = next_update

    def check_refresh(self) -> None:
        ''' Call this method periodically to refresh credentials. '''
        self.schedule.tick()

    @abstractmethod
    def _save_credentials(
        self,
        credentials: 'botocore.credentials.Credentials',
        next_update: datetime.datetime
    ) -> None:
        ''' Saves the credentials for Kinesis Video Stream Producer component.

        This method should be implemented in subclasses.

        :param credentials: The credentials to be saved.
        :param next_update: When the next credentials update is expected.
        '''

    def plugin_config(self) -> str: # pylint: disable=no-self-use
        ''' Returns a string that should be included in the ``kvssink`` GStreamer plugin config.'''
        return ''

    def plugin_config_mask(self, plugin_config: str) -> str: # pylint: disable=no-self-use
        ''' Masks credentials for printing in logs. '''
        return plugin_config


class KVSInlineCredentialsHandler(KVSCredentialsHandler):
    ''' Provides AWS credentials inline in the ``kvssink`` plugin config.

    This credentials handler can be used only with static credentials as there is
    no way to refresh the credentials once they were passed to KVS Producer.

    Args:
        args: Positional arguments to be passed to :class:`KVSCredentialsHandler` initializer.
        kwargs: Keyword arguments to be passed to :class:`KVSCredentialsHandler` initializer.
    '''

    def __init__(self, *args: List[Any], **kwargs: Dict[str, Any]):
        super().__init__(*args, **kwargs)
        if _is_refreshable(self.credentials):
            raise RuntimeError(
                'InlineCredentialsHandler must not be used with refreshable '
                'credentials as they will not be refreshed once passed to '
                'kvssink plugin configuration.'
            )

    def plugin_config(self) -> str:
        return ' '.join([
            f'access-key="{self.credentials.access_key}"',
            f'secret-key="{self.credentials.secret_key}"',
        ])

    def plugin_config_mask(self, plugin_config: str) -> str:
        ''' Masks credentials for printing in logs.

        Args:
            plugin_config: The original plugin config string.
        '''
        res = plugin_config
        res = re.sub(r'secret-key="([^"]*)"', 'secret-key="*****"', res)
        res = re.sub(r'access-key="([^"]*)"', 'access-key="*****"', res)
        return res

    def _save_credentials(self, *args, **kwargs) -> None:
        # Empty implementation, credentials are passed in plugin config
        pass


class KVSEnvironmentCredentialsHandler(KVSCredentialsHandler):
    ''' Saves AWS credentials in environment variables.

    Experience shows that the kvssink plugin does not check periodically the
    environment variables for updated dynamic credentials, so there are chances
    that your stream will stop once the original dynamic credentials expiry.

    For this reason, it is recommened that you use this credentials handler
    only with static credentials and do not use it in production setting.

    Args:
        args: Positional arguments to be passed to :class:`KVSCredentialsHandler` initializer.
        kwargs: Keyword arguments to be passed to :class:`KVSCredentialsHandler` initializer.
    '''
    def __init__(self, *args: List[Any], **kwargs: Dict[str, Any]):
        super().__init__(*args, **kwargs)

    def _save_credentials(
        self,
        credentials: 'botocore.credentials.Credentials',
        _: datetime.datetime
    ):
        if _is_refreshable(credentials):
            os.environ['AWS_SESSION_TOKEN'] = credentials.token
        os.environ['AWS_ACCESS_KEY_ID'] = credentials.access_key
        os.environ['AWS_SECRET_ACCESS_KEY'] = credentials.secret_key


class KVSFileCredentialsHandler(KVSCredentialsHandler):
    ''' Saves AWS variables in a text file compatible with `credential-path`
    parameter of kvssink.

    Using this credentials handler establishes the most robust communication
    between the handler and the Kinesis Video Stream Producer. The credentials
    are written into a text file with a predefined format. This handler can be
    used both with static and dynamic credentials: the handler will ensure that
    the refreshed credentials are updated also in the text file, before the
    declared file expiration.

    Args:
        credentials_path: The path of the credentials file.
        args: Positional arguments to be passed to :class:`KVSCredentialsHandler` initializer.
        kwargs: Keyword arguments to be passed to :class:`KVSCredentialsHandler` initializer.
    '''

    FILE_REFRESH_GRACE_PERIOD = datetime.timedelta(seconds=60)
    '''  Give some time (60s) to boto3 credentials handler to effectively
    refresh the credentials, and declare the expiration of the KVS
    credentials file after the waiting this time, KVS will find the
    new tokens in the file. '''

    def __init__(
        self, credentials_path: str='/tmp/credentials.txt',
        *args: List[Any],
        **kwargs: Dict[str, Any]
    ):
        self.credentials_path = credentials_path
        super().__init__(*args, **kwargs)

    def _save_credentials(self, credentials, next_update):
        if next_update is not None:
            file_expire = next_update + self.FILE_REFRESH_GRACE_PERIOD
            file_expire_str = _format_time(file_expire)
            self.logger.info(f'Credentials file expiration: {file_expire_str}')

            credentials_str = '\t'.join([
                'CREDENTIALS',
                credentials.access_key,
                file_expire_str,
                credentials.secret_key,
                credentials.token
            ])
            self.logger.info(
                f'Updated {self.credentials_path}:\n' + '\t'.join([
                    'CREDENTIALS',
                    f'{credentials.access_key[:5]}...',
                    file_expire_str,
                    f'{credentials.secret_key[:5]}...',
                    f'{credentials.token[:5]}...',
                ])
            )
        else:
            credentials_str = '\t'.join([
                'CREDENTIALS',
                credentials.access_key,
                credentials.secret_key
            ])
            self.logger.info(
                f'Updated {self.credentials_path}:\n' + '\t'.join([
                    'CREDENTIALS',
                    f'{credentials.access_key[:5]}...',
                    f'{credentials.secret_key[:5]}...',
                ])
            )

        with open(self.credentials_path, 'w', encoding='utf-8') as credentials_file:
            credentials_file.write(credentials_str)

    def plugin_config(self) -> str:
        return f'credential-path="{self.credentials_path}"'



from typing import Callable, Tuple, Any

import cv2
import numpy as np

from ..geometry import Point
from .driver import AnnotationDriverBase
from .annotation import (
    MarkerAnnotation, RectAnnotation, LabelAnnotation, LineAnnotation, PolyLineAnnotation
)
from .color import Color

class OpenCVImageAnnotationDriver(AnnotationDriverBase):
    ''' Annotation driver implementation for OpenCV images.

    You should pass an :class:`numpy.ndarray` instance as the context argument of the
    :meth:`~backpack.annotation.AnnotationDriverBase.render()` method. '''

    DEFAULT_COLOR = Color(255, 255, 255)
    DEFAULT_LINEWIDTH = 1
    DEFAULT_FONT = cv2.FONT_HERSHEY_PLAIN

    IMG_HEIGHT_FOR_UNIT_FONT_SCALE = 400
    ''' The height of the image where 1.0 font_scale will be used. '''

    DEFAULT_TEXT_PADDING = (2, 2)

    MARKER_STYLE_TO_CV2 = {
        MarkerAnnotation.Style.CROSS: cv2.MARKER_DIAMOND,
        MarkerAnnotation.Style.TILTED_CROSS: cv2.MARKER_TILTED_CROSS,
        MarkerAnnotation.Style.STAR: cv2.MARKER_STAR,
        MarkerAnnotation.Style.DIAMOND: cv2.MARKER_DIAMOND,
        MarkerAnnotation.Style.SQUARE: cv2.MARKER_SQUARE,
        MarkerAnnotation.Style.TRIANGLE_UP: cv2.MARKER_TRIANGLE_UP,
        MarkerAnnotation.Style.TRIANGLE_DOWN: cv2.MARKER_TRIANGLE_DOWN,
    }

    def draw_transparent(self,
        alpha: float,
        context: np.ndarray,
        drawer: Callable[[np.ndarray], None]
    ) -> None:
        ''' Semi-transparent drawing.

        Args:
            alpha: The transparency factor. Set 0 for opaque drawing, 1 for complete transparency.
            context: The drawing context.
            drawer: A callable that will draw an opaque drawing on the passed-in context. This
                method will make this drawing semi-transparent and render it on the context.
        '''
        if alpha == 1.0:
            drawer(context)
        elif alpha == 0.0:
            return
        else:
            overlay = context.copy()
            drawer(overlay)
            result = cv2.addWeighted(overlay, alpha, context, 1 - alpha, 0)
            np.copyto(context, result)

    @staticmethod
    def scale(point: Any, context: np.ndarray) -> Tuple[float, float]:
        ''' Converts and scales a point instance to an image context '''
        x, y = Point.from_value(point)
        return (int(x * context.shape[1]), int(y * context.shape[0]))

    @staticmethod
    def _color_to_cv2(color: Color) -> Tuple[int, int, int]:
        return (color.b, color.g, color.r)

    def add_rect(self, anno: RectAnnotation, context: np.ndarray) -> None:
        color = anno.color or OpenCVImageAnnotationDriver.DEFAULT_COLOR
        drawer = lambda context: \
            cv2.rectangle(
                img=context,
                pt1=OpenCVImageAnnotationDriver.scale(anno.rect.pt_min, context),
                pt2=OpenCVImageAnnotationDriver.scale(anno.rect.pt_max, context),
                color=OpenCVImageAnnotationDriver._color_to_cv2(color),
                thickness=OpenCVImageAnnotationDriver.DEFAULT_LINEWIDTH
            )
        self.draw_transparent(color.alpha, context, drawer)


    @staticmethod
    def _get_anchor_shift(
        horizontal_anchor: LabelAnnotation.HorizontalAnchor,
        vertical_anchor: LabelAnnotation.VerticalAnchor,
        size_x: int,
        size_y: int,
        baseline: int
    ) -> Tuple[int, int]:
        ''' Gets the shift of the text position based on anchor point location and size of the
        text. '''
        padding_x, padding_y = OpenCVImageAnnotationDriver.DEFAULT_TEXT_PADDING
        shift_x, shift_y = (padding_x, -padding_y)
        if horizontal_anchor == LabelAnnotation.HorizontalAnchor.CENTER:
            shift_x = -size_x / 2
        elif horizontal_anchor == LabelAnnotation.HorizontalAnchor.RIGHT:
            shift_x = -(size_x + padding_x)
        if vertical_anchor == LabelAnnotation.VerticalAnchor.CENTER:
            shift_y = size_y / 2
        elif vertical_anchor == LabelAnnotation.VerticalAnchor.TOP:
            shift_y = size_y + padding_y
        elif vertical_anchor == LabelAnnotation.VerticalAnchor.BASELINE:
            shift_y = baseline
        return (int(shift_x), int(shift_y))

    def add_label(self, anno: LabelAnnotation, context: np.ndarray) -> None:
        ctx_height = context.shape[0]
        scale = ctx_height / OpenCVImageAnnotationDriver.IMG_HEIGHT_FOR_UNIT_FONT_SCALE * anno.size
        thickness = max(int(scale), 1)
        font = OpenCVImageAnnotationDriver.DEFAULT_FONT
        x, y = OpenCVImageAnnotationDriver.scale(anno.point, context)
        color = anno.color or OpenCVImageAnnotationDriver.DEFAULT_COLOR
        if (anno.horizontal_anchor != LabelAnnotation.HorizontalAnchor.LEFT or
            anno.vertical_anchor != LabelAnnotation.VerticalAnchor.BOTTOM):
            (size_x, size_y), baseline = cv2.getTextSize(
                text=anno.text, fontFace=font, fontScale=scale, thickness=thickness
            )
            shift_x, shift_y = OpenCVImageAnnotationDriver._get_anchor_shift(
                anno.horizontal_anchor,
                anno.vertical_anchor,
                size_x, size_y, baseline
            )
            x += shift_x
            y += shift_y

        drawer = lambda context: \
            cv2.putText(
                img=context,
                text=anno.text,
                org=(x, y),
                fontFace=font,
                fontScale=scale,
                color=OpenCVImageAnnotationDriver._color_to_cv2(color),
                thickness=thickness
            )
        self.draw_transparent(color.alpha, context, drawer)

    def add_marker(self, anno: MarkerAnnotation, context: np.ndarray) -> None:
        markerType = OpenCVImageAnnotationDriver.MARKER_STYLE_TO_CV2.get(
            anno.style, cv2.MARKER_DIAMOND
        )
        color = anno.color or OpenCVImageAnnotationDriver.DEFAULT_COLOR
        drawer = lambda context: \
            cv2.drawMarker(
                img=context,
                position=OpenCVImageAnnotationDriver.scale(anno.point, context),
                color=OpenCVImageAnnotationDriver._color_to_cv2(color),
                markerType=markerType
            )
        self.draw_transparent(color.alpha, context, drawer)

    def add_line(self, anno: LineAnnotation, context: Any) -> None:
        color = anno.color or OpenCVImageAnnotationDriver.DEFAULT_COLOR
        drawer = lambda context: \
            cv2.line(
                img=context,
                pt1=OpenCVImageAnnotationDriver.scale(anno.line.pt1, context),
                pt2=OpenCVImageAnnotationDriver.scale(anno.line.pt2, context),
                color=OpenCVImageAnnotationDriver._color_to_cv2(color),
                thickness=anno.thickness
            )
        self.draw_transparent(color.alpha, context, drawer)

    def add_polyline(self, anno: PolyLineAnnotation, context: Any) -> None:
        pts = [OpenCVImageAnnotationDriver.scale(pt, context) for pt in anno.polyline.points]
        pts = [np.array(pts, dtype=np.int32)]
        if anno.fill_color is not None:
            fill_color = OpenCVImageAnnotationDriver._color_to_cv2(anno.fill_color)
            if anno.polyline.is_convex:
                drawer = lambda context: \
                    cv2.fillConvexPoly(img=context, points=pts[0], color=fill_color)
            else:
                drawer = lambda context: \
                    cv2.fillPoly(img=context, pts=pts, color=fill_color)
            self.draw_transparent(anno.fill_color.alpha, context, drawer)

        color = anno.color or OpenCVImageAnnotationDriver.DEFAULT_COLOR
        drawer = lambda context: \
            cv2.polylines(
                img=context,
                pts=pts,
                isClosed=anno.polyline.closed,
                color=OpenCVImageAnnotationDriver._color_to_cv2(color),
                thickness=anno.thickness
            )
        self.draw_transparent(color.alpha, context, drawer)



'''
This module contains RTSPSkyLine and RTSPServer. These two classes allow serving a sequence of
OpenCV images as video streams using the RTSP protocol.

To use this class you MUST have the following dependencies correctly configured on your system:

 - `GStreamer 1.0`_ installed with standard plugins pack, libav, tools and development libraries
 - `OpenCV 4.2.0`_, compiled with GStreamer support and Python bindings
 - `gst-rtsp-server`_ with development libraries (libgstrtspserver-1.0-dev)

These dependencies can not be easily specified by a ``requirements.txt`` or a Conda environment.
See the example ``Dockerfile`` on how to install these dependencies on your system.

.. _`GStreamer 1.0`: https://gstreamer.freedesktop.org
.. _`OpenCV 4.2.0`: https://opencv.org/opencv-4-2-0/
.. _`gst-rtsp-server`: https://github.com/GStreamer/gst-rtsp-server
'''

from typing import Any, Optional, List
import threading
import logging

import gi

gi.require_version('Gst', '1.0')
gi.require_version('GstRtspServer', '1.0')
from gi.repository import GLib, GstRtspServer

from .skyline import SkyLine

class RTSPServer:
    ''' The :class:`RTSPServer` instance wraps a GStreamer RTSP server that serves video streams
    to clients using the RTSP protocol.

    You typically want to have a single instance of :class:`RTSPServer` for your application.
    You can register any number of video streams that will be served by a single instance of
    RTSP server. The port number of the server should be unique among all applications on the
    device.

    For an example usage of :class:`RTSPServer`, see the documentation of :class:`RTSPSkyLine`
    class.

    Args:
        port: The port to listen on. You can not modify the port after the initialization of
            the :class:`RTSPServer` instance. Defaults to ``8554``.
        parent_logger: If you want to connect the logger of :class:`RTSPServer` to a parent,
            specify it here.
    '''

    def __init__(
        self,
        port: str = '8554',
        parent_logger: Optional[logging.Logger] = None,
    ):
        self.logger = (
            logging.getLogger(self.__class__.__name__) if parent_logger is None else
            parent_logger.getChild(self.__class__.__name__)
        )
        self.gst_server = GstRtspServer.RTSPServer()
        self.gst_server.service = port
        self.streams = {}
        self._port = port
        self._loop = GLib.MainLoop()
        self._thread = None

    def add_stream(self, mount_point: str, pipeline: str) -> None:
        ''' Registers a new video stream to the server.

        Args:
            mount_point: The path that will be used to access the stream. For example,
                if you specify ``/my_stream``, the stream will be accessible for clients using
                the ``rtsp://127.0.0.1:8854/mystream`` url (change the IP address and the port
                number accordingly).
            pipeline: The GStreamer pipeline to use for the stream. This will be typically
                a pipeline picking up the raw UDP packets from a local port and wrapping it to a
                H.264 envelope, for example:
                ``udpsrc port=5000 ! application/x-rtp,media=video,encoding-name=H264 ! rtph264depay !
                rtph264pay name=pay0``
        '''
        self.logger.info('Adding pipeline to mount point "%s"', mount_point)
        mounts = self.gst_server.get_mount_points()
        factory = GstRtspServer.RTSPMediaFactory()
        factory.set_launch(pipeline)
        factory.set_shared(True)
        mounts.add_factory(mount_point, factory)
        self.streams[mount_point] = pipeline

    def remove_stream(self, mount_point: str) -> None:
        ''' It removes a registered stream from the server.

        :param mount_point: The registered path of the stream.
        '''
        mounts = self.gst_server.get_mount_points()
        mounts.remove_factory(mount_point)
        del self.streams[mount_point]

    def start(self):
        ''' Starts the RTSP server asynchronously.

        After calling this method, the RTSP requests will be served.
        '''
        self.logger.info('Starting RTSPServer')
        self.gst_server.attach()
        self._thread = threading.Thread(
            name='RTSPServerThread',
            target=self._loop.run,
            daemon=True
        )
        self._thread.start()

    def stop(self):
        ''' Stops the RTSP server.

        After calling this method, no RTSP requests will be served. The server can be restarted
        later on calling the :meth:`start()` method.
        '''
        self._loop.quit()

    @property
    def port(self) -> str:
        ''' The port where this server listents to incoming connections. '''
        return self._port

    def urls(self) -> List[str]:
        ''' Returns the list of URLs where the server serves RTSP streams. '''
        return [f'rtsp://127.0.0.1:{self.port}{mp}' for mp in self.streams.keys()]

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} streams=[{", ".join(self.urls())}]>'



class RTSPSkyLine(SkyLine):
    ''' Together with :class:`RTSPServer`, :class:`RTSPSkyLine` a sequence of OpenCV frames
    on the RTSP protocol.

    A single instance of :class:`RTSPServer` application can serve streams coming from multiple
    :class:`RTSPSkyLine` instances. You should instantiate the :class:`RTSPServer` instance first.
    For example, if you want to serve two seperate RTSP streams, you could use this code to set up
    your scenario::

        server = RTSPServer(port="8554")
        skyline1 = RTSPSkyLine(server, "/stream1")
        skyline2 = RTSPSkyLine(server, "/stream2")
        skyline1.start_streaming(30, 640, 480)
        skyline2.start_streaming(30, 640, 480)
        server.start()

        while True:
            frame1 = ... # Get frame for the first stream as a numpy array of shape (640, 480, 3)
            frame2 = ... # Get frame for the second stream
            skyline1.put(frame1)
            skyline2.put(frame2)

    Using this code, you can access the streams at the following URLs:

        - ``rtsp://127.0.0.1:8554/stream1``
        - ``rtsp://127.0.0.1:8554/stream2``

    If the application (or the firewall) is configured to allow incoming connections on the ``8554``
    port, the streams will be accessibly also from the external ip of the device.

    Args:
        server: The RTSPServer instance that this stream is being served by
        path: The path to the stream. This is the path that the client will use to connect to
            the stream
        args: Positional arguments to be passed to :class:`~backpack.skyline.SkyLine`
            superclass initializer.
        kwargs: Keyword arguments to be passed to :class:`~backpack.skyline.SkyLine`
            superclass initializer.
    '''

    LOCALHOST = '127.0.0.1'
    last_loopback_port = 5000

    def __init__(self, server: RTSPServer, path: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.server = server
        self.path = path if path.startswith('/') else '/' + path
        self.loopback_port = RTSPSkyLine.last_loopback_port
        RTSPSkyLine.last_loopback_port += 1
        self.server.add_stream(self.path, self._get_server_pipeline())

    def _get_pipeline(self, fps: float, width: int, height: int) -> str:
        pipeline = ' ! '.join([
            'appsrc',
            'queue',
            'videoconvert',
            f'video/x-raw,format=I420,width={width},height={height},framerate={fps}/1',
            'x264enc bframes=0 key-int-max=45 bitrate=500',
            'video/x-h264,stream-format=avc,alignment=au,profile=baseline',
            'h264parse',
            'rtph264pay',
            f'udpsink host={self.LOCALHOST} port={self.loopback_port}'
        ])
        self.logger.info(f'GStreamer application pipeline definition:\n{pipeline}')
        return pipeline

    def _get_server_pipeline(self):
        pipeline = ' ! '.join([
            f'udpsrc port={self.loopback_port}',
            'application/x-rtp,media=video,encoding-name=H264',
            'rtph264depay',
            'rtph264pay name=pay0'
        ])
        self.logger.info(f'GStreamer RTSP server pipeline definition:\n{pipeline}')
        return pipeline



from glob import glob
import os

from setuptools import setup, find_packages

name = "panorama-backpack"
version = "0.1.5"

# get the dependencies and installs
with open("requirements.txt", "r") as f:
    requires = [x.strip() for x in f if x.strip()]

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name=name,
    version=version,
    description="Tools for AWS Panorama development",
    python_requires=">=3.7",
    packages=find_packages(exclude=["tests.*", "tests"]),
    package_dir={"backpack": "backpack"},
    include_package_data=True,
    install_requires=requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Janos Tolgyesi",
    author_email='janos.tolgyesi@neosperience.com',
    url="https://github.com/Neosperience/backpack",
    download_url="https://github.com/Neosperience/backpack/archive/refs/tags/v{}.tar.gz".format(version),
    keywords=["aws", "panorama", "video-analysis"],
    entry_points={},
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3.7",
    ],
    extras_require={
    },
)



''' This module makes it possible to draw annotations on different
backends with an unified API. Currently, you can draw rectangles and labels with
:mod:`~backpack.annotation` on ``panoramasdk.media`` and OpenCV images
(:class:`numpy arrays <numpy.ndarray>`).'''

from typing import Optional
from enum import Enum
import datetime
from abc import ABC
from dataclasses import dataclass

from .color import Color
from ..timepiece import local_now
from ..geometry import Point, Rectangle, Line, PolyLine


class Annotation(ABC):
    ''' Abstract base class for annotations. '''


@dataclass(frozen=True)
class LabelAnnotation(Annotation):
    ''' A label annotation to be rendered in an :class:`AnnotationDriver` context.

    The :attr:`point` refers to the position of the anchor point.

    Args:
        point (Point): The origin of the label
        text (str): The text to be rendered
        color (Color): The color of the text
    '''
    class HorizontalAnchor(Enum):
        ''' Horizontal anchor point location. '''
        LEFT = 1,
        ''' Left anchor point '''
        CENTER = 2,
        ''' Center anchor point '''
        RIGHT = 3
        ''' Right anchor point '''

    class VerticalAnchor(Enum):
        ''' Vertical anchor point location. '''
        TOP = 1
        ''' Top anchor point '''
        CENTER = 2
        ''' Center anchor point '''
        BASELINE = 3
        ''' Text baseline anchor point '''
        BOTTOM = 4
        ''' Bottom anchor point '''

    point: Point
    ''' The origin of the label. '''

    text: str
    ''' The text to be rendered. '''

    color: Color = None
    ''' The color of the text. If `None`, the default drawing color will be used. '''

    horizontal_anchor: HorizontalAnchor = HorizontalAnchor.LEFT
    ''' Horizontal anchor point location '''

    vertical_anchor: VerticalAnchor = VerticalAnchor.BOTTOM
    ''' Vertical anchor point location '''

    size: float = 1.0
    ''' The relative size of the text '''


@dataclass(frozen=True)
class RectAnnotation(Annotation):
    ''' A rectangle annotation to be rendered in an AnnotationDriver context.

    Args:
        rect: The rectangle to be rendered in the AnnotationDriver context.
        color: The line color of the rectangle.
    '''
    rect: Rectangle
    ''' The rectangle to be rendered in the AnnotationDriver context '''

    color: Color = None
    ''' The line color of the rectangle. If `None`, the default drawing color will be used. '''


@dataclass(frozen=True)
class LineAnnotation(Annotation):
    ''' A line annotation to be rendered in an AnnotationDriver context.

    Args:
        line: The line segment to be drawn
        color: The color of the line
    '''
    line: Line
    ''' The line segment to be drawn '''

    color: Color = None
    ''' The color of the line. If `None`, the default drawing color will be used. '''

    thickness: int = 1
    ''' The thickness of the line. '''


@dataclass(frozen=True)
class MarkerAnnotation(Annotation):
    ''' A marker annotation to be rendered in an AnnotationDriver context.

    Args:
        point (Point): The coordinates of the maker
        style (MarkerAnnotation.Style): The style of the marker
        color (Color): The color of the maker
    '''
    class Style(Enum):
        ''' Possible set of marker types. '''
        CROSS = 1
        ''' A crosshair marker shape. '''
        TILTED_CROSS = 2
        ''' A 45 degree tilted crosshair marker shape. '''
        STAR = 3
        ''' A star marker shape, combination of cross and tilted cross. '''
        DIAMOND = 4
        ''' A diamond marker shape. '''
        SQUARE = 5
        ''' A square marker shape. '''
        TRIANGLE_UP = 6
        ''' An upwards pointing triangle marker shape. '''
        TRIANGLE_DOWN = 7
        ''' A downwards pointing triangle marker shape. '''

    point: Point
    ''' The coordinates of the maker. '''
    style: Style = Style.CROSS
    ''' The style of the marker. '''
    color: Color = None
    ''' The color of the maker. '''


@dataclass(frozen=True)
class PolyLineAnnotation(Annotation):
    ''' A PolyLine annotation to be rendered in an AnnotationDriver context.

    Args:
        polyline(PolyLine): The PolyLine instance
        thickness(int): Line thickness
        color(Color): Line color
        fill_color(Color): Fill color
    '''

    polyline : PolyLine
    ''' The PolyLine instance. '''
    thickness : int = 1
    ''' Line thickness. '''
    color : Color = None
    ''' Line color. '''
    fill_color : Color = None
    ''' Fill color. '''


@dataclass(frozen=True)
class TimestampAnnotation(LabelAnnotation):
    ''' A timestamp annotation to be rendered in an AnnotationDriver context.

    Args:
        timestamp: The timestamp to be rendered. If not specified, the current local time
            will be used.
        point: The origin of the timestamp. If not specified, the timestamp will be places
            on the top-left corner of the image.
    '''
    def __init__(self, timestamp: Optional[datetime.datetime]=None, point: Point=Point(0.02, 0.04)):
        timestamp = timestamp or local_now()
        time_str = timestamp.strftime('%Y-%m-%d %H:%M:%S')
        return super().__init__(point=point, text=time_str)


@dataclass(frozen=True)
class BoundingBoxAnnotation(Annotation):
    ''' A bounding box annotation with a rectangle, and optional upper and lower labels.

    Args:
        rectangle: The rectangle of the bounding box.
        top_label: The optional top label.
        bottom_label: The optional bottom label.
        color: The color of the bounding box and the labels.
    '''

    rectangle: Rectangle
    ''' The rectangle of the bounding box. '''

    top_label: Optional[str] = None
    ''' The optional top label. '''

    bottom_label: Optional[str] = None
    ''' The optional bottom label. '''

    color: Optional[Color] = None
    ''' The color of the bounding box and the labels. '''
